/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2014 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.kiml.formats.gml;

import com.google.common.base.Objects;
import de.cau.cs.kieler.kiml.formats.gml.CollectionElement;
import de.cau.cs.kieler.kiml.formats.gml.Element;
import de.cau.cs.kieler.kiml.formats.gml.GMLModel;
import de.cau.cs.kieler.kiml.formats.gml.NumberElement;
import de.cau.cs.kieler.kiml.formats.gml.StringElement;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StreamTokenizer;
import java.util.List;
import java.util.Stack;

/**
 * Parses the GML files in one top-down pass using a {@link StreamTokenizer}.
 * For more information on the Graph Modelling Language see
 * http://www.fim.uni-passau.de/en/fim/faculty/chairs/theoretische-informatik/projects.html.
 * 
 * Note that the StreamTokenizer assumes all numbers to be doubles. To
 * add support for integers we interpret a numbers {@code n} as integer
 * if {@code n % 1 == 0}. For instance, both {@code 1} and {@code 1.0}
 * are converted to integers.
 * 
 * @author uru
 * @see GMLModel
 * @see GMLSerializer
 */
@SuppressWarnings("all")
public final class GMLParser {
  private final static char SBO = '[';
  
  private final static char SBC = ']';
  
  private final static char TT_QUOTED = '\"';
  
  private final static int KEY_OR_CLOSE = 0;
  
  private final static int VAL_OR_COL = 1;
  
  public static GMLModel parse(final InputStream is) throws IOException {
    InputStreamReader _inputStreamReader = new InputStreamReader(is);
    final BufferedReader r = new BufferedReader(_inputStreamReader);
    final StreamTokenizer st = new StreamTokenizer(r);
    st.slashSlashComments(true);
    st.slashStarComments(true);
    int expecting = GMLParser.KEY_OR_CLOSE;
    Element currentElement = null;
    String currentKey = "";
    final Stack<CollectionElement> elementStack = new Stack<CollectionElement>();
    CollectionElement _collectionElement = new CollectionElement(null, "root");
    elementStack.push(_collectionElement);
    boolean eof = false;
    do {
      {
        final int token = st.nextToken();
        if ((expecting == GMLParser.KEY_OR_CLOSE)) {
          switch (token) {
            case StreamTokenizer.TT_EOF:
              eof = true;
              break;
            case StreamTokenizer.TT_EOL:
              break;
            case StreamTokenizer.TT_WORD:
              currentKey = st.sval;
              expecting = GMLParser.VAL_OR_COL;
              break;
            default:
              if ((((char) token) == GMLParser.SBC)) {
                int _size = elementStack.size();
                boolean _equals = (_size == 1);
                if (_equals) {
                  throw new IllegalArgumentException("Too many closing \']\'.");
                }
                CollectionElement _pop = elementStack.pop();
                currentElement = _pop;
                expecting = GMLParser.KEY_OR_CLOSE;
              } else {
                String _plus = (("Expected key or \']\' but read \'" + Character.valueOf(((char) token))) + "\' at line ");
                int _lineno = st.lineno();
                String _plus_1 = (_plus + Integer.valueOf(_lineno));
                String _plus_2 = (_plus_1 + ".");
                throw new IllegalArgumentException(_plus_2);
              }
              break;
          }
        } else {
          if ((expecting == GMLParser.VAL_OR_COL)) {
            final CollectionElement container = elementStack.peek();
            Element newElement = null;
            switch (token) {
              case StreamTokenizer.TT_EOF:
                eof = true;
                break;
              case StreamTokenizer.TT_EOL:
                break;
              case StreamTokenizer.TT_WORD:
              case GMLParser.TT_QUOTED:
                StringElement _stringElement = new StringElement(container, currentKey, st.sval);
                newElement = _stringElement;
                expecting = GMLParser.KEY_OR_CLOSE;
                break;
              case StreamTokenizer.TT_NUMBER:
                Number _xifexpression = null;
                if (((st.nval % 1) == 0.0)) {
                  int _intValue = Double.valueOf(st.nval).intValue();
                  _xifexpression = ((Integer) Integer.valueOf(_intValue));
                } else {
                  _xifexpression = ((Double) Double.valueOf(st.nval));
                }
                final Number n = ((Number)_xifexpression);
                NumberElement _numberElement = new NumberElement(container, currentKey, n);
                newElement = _numberElement;
                expecting = GMLParser.KEY_OR_CLOSE;
                break;
              default:
                if ((((char) token) == GMLParser.SBO)) {
                  CollectionElement _collectionElement_1 = new CollectionElement(container, currentKey);
                  newElement = _collectionElement_1;
                  expecting = GMLParser.KEY_OR_CLOSE;
                } else {
                  String _plus = (("Expected value or but read " + Character.valueOf(((char) token))) + "\' at line ");
                  int _lineno = st.lineno();
                  String _plus_1 = (_plus + Integer.valueOf(_lineno));
                  String _plus_2 = (_plus_1 + ".");
                  throw new IllegalArgumentException(_plus_2);
                }
                break;
            }
            boolean _notEquals = (!Objects.equal(newElement, null));
            if (_notEquals) {
              List<Element> _elements = container.getElements();
              _elements.add(newElement);
              if ((newElement instanceof CollectionElement)) {
                elementStack.push(((CollectionElement) newElement));
              }
              currentElement = newElement;
            }
          }
        }
      }
    } while((!eof));
    int _size = elementStack.size();
    boolean _lessThan = (_size < 1);
    if (_lessThan) {
      throw new IllegalArgumentException("Too many closing \']\'.");
    } else {
      int _size_1 = elementStack.size();
      boolean _greaterThan = (_size_1 > 1);
      if (_greaterThan) {
        throw new IllegalArgumentException("Too many opening \'[\'.");
      }
    }
    final CollectionElement root = elementStack.pop();
    final GMLModel model = new GMLModel();
    List<Element> _elements = model.getElements();
    List<Element> _elements_1 = root.getElements();
    _elements.addAll(_elements_1);
    return model;
  }
}
