/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.klighd.examples.ecore;

import com.google.common.base.Objects;
import com.google.common.base.Strings;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import de.cau.cs.kieler.klighd.KlighdConstants;
import de.cau.cs.kieler.klighd.SynthesisOption;
import de.cau.cs.kieler.klighd.examples.ecore.EModelElementCollection;
import de.cau.cs.kieler.klighd.krendering.Colors;
import de.cau.cs.kieler.klighd.krendering.KAreaPlacementData;
import de.cau.cs.kieler.klighd.krendering.KContainerRendering;
import de.cau.cs.kieler.klighd.krendering.KGridPlacement;
import de.cau.cs.kieler.klighd.krendering.KGridPlacementData;
import de.cau.cs.kieler.klighd.krendering.KImage;
import de.cau.cs.kieler.klighd.krendering.KPolygon;
import de.cau.cs.kieler.klighd.krendering.KPolyline;
import de.cau.cs.kieler.klighd.krendering.KPosition;
import de.cau.cs.kieler.klighd.krendering.KRectangle;
import de.cau.cs.kieler.klighd.krendering.KRendering;
import de.cau.cs.kieler.klighd.krendering.KText;
import de.cau.cs.kieler.klighd.krendering.extensions.KContainerRenderingExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KEdgeExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KNodeExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KPolylineExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KRenderingExtensions;
import de.cau.cs.kieler.klighd.syntheses.AbstractDiagramSynthesis;
import de.cau.cs.kieler.klighd.util.KlighdProperties;
import de.cau.cs.kieler.klighd.util.KlighdSemanticDiagramData;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.function.Consumer;
import javax.inject.Inject;
import org.eclipse.elk.core.klayoutdata.KLayoutData;
import org.eclipse.elk.core.options.CoreOptions;
import org.eclipse.elk.core.options.Direction;
import org.eclipse.elk.core.options.EdgeType;
import org.eclipse.elk.core.util.Pair;
import org.eclipse.elk.graph.KEdge;
import org.eclipse.elk.graph.KNode;
import org.eclipse.elk.graph.properties.IProperty;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EEnumLiteral;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IntegerRange;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ListExtensions;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * This diagram synthesis implementation demonstrates the usage of KLighD for the purpose of
 * representing content of models by means of graph-like diagrams.
 * 
 * This translation provides the following synthesis options:
 * <ol>
 *  <li>Depicting the selected classes only.</li>
 *  <li>Depicting the selected classes, and directly related ones.</li>
 *  <li>Depicting all classes of the Ecore model, the selected ones are highlighted.</li>
 *  <li>Depicting the attributes of the classes</li>
 * <ol>
 */
@SuppressWarnings("all")
public class EcoreDiagramSynthesis extends AbstractDiagramSynthesis<EModelElementCollection> {
  @Inject
  @Extension
  private KNodeExtensions _kNodeExtensions;
  
  @Inject
  @Extension
  private KEdgeExtensions _kEdgeExtensions;
  
  @Inject
  @Extension
  private KRenderingExtensions _kRenderingExtensions;
  
  @Inject
  @Extension
  private KContainerRenderingExtensions _kContainerRenderingExtensions;
  
  @Inject
  @Extension
  private KPolylineExtensions _kPolylineExtensions;
  
  private final static String CHOSEN = "Chosen classes";
  
  private final static String CHOSEN_AND_RELATED = "Chosen (highlighted) && related classes";
  
  private final static String ALL = "All classes, selection highlighted";
  
  private final static String CLASS_FILTER_NAME = "Class filter";
  
  private final static Colors LINE_COLOR = Colors.GRAY_25;
  
  /**
   * The filter option definition that allows users to customize the constructed class diagrams.
   */
  private final static SynthesisOption CLASS_FILTER = SynthesisOption.createChoiceOption(EcoreDiagramSynthesis.CLASS_FILTER_NAME, 
    ImmutableList.<String>of(EcoreDiagramSynthesis.CHOSEN, EcoreDiagramSynthesis.CHOSEN_AND_RELATED, EcoreDiagramSynthesis.ALL), EcoreDiagramSynthesis.CHOSEN_AND_RELATED);
  
  /**
   * Option to activate/deactivate the attribute lists.
   */
  private final static SynthesisOption RELATED_CLASSES_DEPTH = SynthesisOption.<Integer>createRangeOption("Reference depth", Integer.valueOf(1), Integer.valueOf(3), Integer.valueOf(1), Integer.valueOf(1));
  
  /**
   * Option choose the reference depth while determining the classes related to the selected ones.
   */
  private final static SynthesisOption ATTRIBUTES = SynthesisOption.createCheckOption("Attributes/Literals", Boolean.valueOf(false));
  
  /**
   * {@inheritDoc}<br>
   * <br>
   * Registers the diagram filter option declared above, which allow users to tailor the constructed diagrams.
   */
  @Override
  public List<SynthesisOption> getDisplayedSynthesisOptions() {
    return ImmutableList.<SynthesisOption>of(EcoreDiagramSynthesis.CLASS_FILTER, EcoreDiagramSynthesis.RELATED_CLASSES_DEPTH, EcoreDiagramSynthesis.ATTRIBUTES);
  }
  
  /**
   * {@inheritDoc}<br>
   * <br>
   * Registers reasonable layout options that are recommended to users to tailor the constructed diagram layouts.
   */
  @Override
  public List<Pair<IProperty<?>, List<?>>> getDisplayedLayoutOptions() {
    Direction[] _values = Direction.values();
    Iterable<Direction> _drop = IterableExtensions.<Direction>drop(((Iterable<Direction>)Conversions.doWrapArray(_values)), 1);
    final Function1<Direction, String> _function = new Function1<Direction, String>() {
      @Override
      public String apply(final Direction it) {
        return it.name();
      }
    };
    List<Direction> _sortBy = IterableExtensions.<Direction, String>sortBy(_drop, _function);
    Pair<IProperty<?>, List<?>> _specifyLayoutOption = this.specifyLayoutOption(CoreOptions.DIRECTION, _sortBy);
    ArrayList<Integer> _newArrayList = CollectionLiterals.<Integer>newArrayList(Integer.valueOf(0), Integer.valueOf(255));
    Pair<IProperty<?>, List<?>> _specifyLayoutOption_1 = this.specifyLayoutOption(CoreOptions.SPACING_NODE, _newArrayList);
    return ImmutableList.<Pair<IProperty<?>, List<?>>>of(_specifyLayoutOption, _specifyLayoutOption_1);
  }
  
  /**
   * {@inheritDoc}<br>
   * <br>
   * This main method creates the root node that represents the canvas of the diagram.
   * It configures some layout options adds the diagram elements by distinguishing the three option cases.
   * 
   * Note that this translation added also the Classes contained in selected EPackages,
   * see the end of this method.
   */
  @Override
  public KNode transform(final EModelElementCollection choice) {
    KNode _createNode = this._kNodeExtensions.createNode();
    final Procedure1<KNode> _function = new Procedure1<KNode>() {
      @Override
      public void apply(final KNode it) {
        EcoreDiagramSynthesis.this._kNodeExtensions.<String>addLayoutParam(it, CoreOptions.ALGORITHM, "de.cau.cs.kieler.kiml.ogdf.planarization");
        EcoreDiagramSynthesis.this._kNodeExtensions.<Float>addLayoutParam(it, CoreOptions.SPACING_NODE, Float.valueOf(75f));
        EcoreDiagramSynthesis.this._kNodeExtensions.<Direction>addLayoutParam(it, CoreOptions.DIRECTION, Direction.UP);
        Iterable<EClassifier> _filter = Iterables.<EClassifier>filter(choice, EClassifier.class);
        final List<EClassifier> depictedClasses = IterableExtensions.<EClassifier>toList(_filter);
        Object _objectValue = EcoreDiagramSynthesis.this.getObjectValue(EcoreDiagramSynthesis.CLASS_FILTER);
        boolean _equals = Objects.equal(_objectValue, EcoreDiagramSynthesis.CHOSEN);
        if (_equals) {
          EcoreDiagramSynthesis.this.createElementFigures(depictedClasses, it);
        } else {
          Object _objectValue_1 = EcoreDiagramSynthesis.this.getObjectValue(EcoreDiagramSynthesis.CLASS_FILTER);
          boolean _equals_1 = Objects.equal(_objectValue_1, EcoreDiagramSynthesis.CHOSEN_AND_RELATED);
          if (_equals_1) {
            Iterable<EClass> _filter_1 = Iterables.<EClass>filter(choice, EClass.class);
            EClass _head = null;
            if (_filter_1!=null) {
              _head=IterableExtensions.<EClass>head(_filter_1);
            }
            EObject _eContainer = null;
            if (_head!=null) {
              _eContainer=_head.eContainer();
            }
            final EPackage ePackage = ((EPackage) _eContainer);
            final HashSet<EClassifier> classesCopy = Sets.<EClassifier>newHashSet(depictedClasses);
            int _intValue = EcoreDiagramSynthesis.this.getIntValue(EcoreDiagramSynthesis.RELATED_CLASSES_DEPTH);
            IntegerRange _upTo = new IntegerRange(1, _intValue);
            final Consumer<Integer> _function = new Consumer<Integer>() {
              @Override
              public void accept(final Integer it) {
                Iterable<EClass> _filter = Iterables.<EClass>filter(classesCopy, EClass.class);
                final Procedure1<Iterable<EClass>> _function = new Procedure1<Iterable<EClass>>() {
                  @Override
                  public void apply(final Iterable<EClass> it) {
                    final Consumer<EClass> _function = new Consumer<EClass>() {
                      @Override
                      public void accept(final EClass it) {
                        EList<EStructuralFeature> _eStructuralFeatures = it.getEStructuralFeatures();
                        Iterable<EReference> _filter = Iterables.<EReference>filter(_eStructuralFeatures, EReference.class);
                        final Function1<EReference, EClassifier> _function = new Function1<EReference, EClassifier>() {
                          @Override
                          public EClassifier apply(final EReference it) {
                            return it.getEType();
                          }
                        };
                        Iterable<EClassifier> _map = IterableExtensions.<EReference, EClassifier>map(_filter, _function);
                        Iterables.<EClassifier>addAll(depictedClasses, _map);
                        EList<EStructuralFeature> _eStructuralFeatures_1 = it.getEStructuralFeatures();
                        Iterable<EAttribute> _filter_1 = Iterables.<EAttribute>filter(_eStructuralFeatures_1, EAttribute.class);
                        final Function1<EAttribute, EClassifier> _function_1 = new Function1<EAttribute, EClassifier>() {
                          @Override
                          public EClassifier apply(final EAttribute it) {
                            return it.getEType();
                          }
                        };
                        Iterable<EClassifier> _map_1 = IterableExtensions.<EAttribute, EClassifier>map(_filter_1, _function_1);
                        final Function1<EClassifier, Boolean> _function_2 = new Function1<EClassifier, Boolean>() {
                          @Override
                          public Boolean apply(final EClassifier it) {
                            EObject _eContainer = it.eContainer();
                            return Boolean.valueOf(Objects.equal(_eContainer, ePackage));
                          }
                        };
                        Iterable<EClassifier> _filter_2 = IterableExtensions.<EClassifier>filter(_map_1, _function_2);
                        Iterables.<EClassifier>addAll(depictedClasses, _filter_2);
                      }
                    };
                    it.forEach(_function);
                    final Consumer<EClass> _function_1 = new Consumer<EClass>() {
                      @Override
                      public void accept(final EClass it) {
                        EList<EClass> _eSuperTypes = it.getESuperTypes();
                        depictedClasses.addAll(_eSuperTypes);
                      }
                    };
                    it.forEach(_function_1);
                    List<EClassifier> _elvis = null;
                    EList<EClassifier> _eClassifiers = null;
                    if (ePackage!=null) {
                      _eClassifiers=ePackage.getEClassifiers();
                    }
                    if (((List<EClassifier>) _eClassifiers) != null) {
                      _elvis = ((List<EClassifier>) _eClassifiers);
                    } else {
                      List<EClassifier> _emptyList = CollectionLiterals.<EClassifier>emptyList();
                      _elvis = _emptyList;
                    }
                    Iterable<EClass> _filter = Iterables.<EClass>filter(_elvis, EClass.class);
                    final Function1<EClass, Boolean> _function_2 = new Function1<EClass, Boolean>() {
                      @Override
                      public Boolean apply(final EClass it) {
                        boolean _xblockexpression = false;
                        {
                          EList<EClass> _eSuperTypes = it.getESuperTypes();
                          final ArrayList<EClass> sts = Lists.<EClass>newArrayList(((Iterable<EClass>) _eSuperTypes));
                          sts.retainAll(classesCopy);
                          EList<EReference> _eReferences = it.getEReferences();
                          ArrayList<EReference> _newArrayList = Lists.<EReference>newArrayList(((Iterable<EReference>) _eReferences));
                          final Function1<EReference, EClassifier> _function = new Function1<EReference, EClassifier>() {
                            @Override
                            public EClassifier apply(final EReference it) {
                              return it.getEType();
                            }
                          };
                          final List<EClassifier> rts = ListExtensions.<EReference, EClassifier>map(_newArrayList, _function);
                          rts.retainAll(classesCopy);
                          boolean _or = false;
                          boolean _isEmpty = sts.isEmpty();
                          boolean _not = (!_isEmpty);
                          if (_not) {
                            _or = true;
                          } else {
                            boolean _isEmpty_1 = rts.isEmpty();
                            boolean _not_1 = (!_isEmpty_1);
                            _or = _not_1;
                          }
                          _xblockexpression = _or;
                        }
                        return Boolean.valueOf(_xblockexpression);
                      }
                    };
                    Iterable<EClass> _filter_1 = IterableExtensions.<EClass>filter(_filter, _function_2);
                    Iterables.<EClassifier>addAll(depictedClasses, _filter_1);
                  }
                };
                ObjectExtensions.<Iterable<EClass>>operator_doubleArrow(_filter, _function);
                classesCopy.addAll(depictedClasses);
              }
            };
            _upTo.forEach(_function);
            EcoreDiagramSynthesis.this.createElementFigures(depictedClasses, it);
            Iterable<EClass> _filter_2 = Iterables.<EClass>filter(choice, EClass.class);
            final Consumer<EClass> _function_1 = new Consumer<EClass>() {
              @Override
              public void accept(final EClass it) {
                KNode _node = EcoreDiagramSynthesis.this._kNodeExtensions.getNode(it);
                KRendering _kRendering = EcoreDiagramSynthesis.this._kRenderingExtensions.getKRendering(_node);
                EcoreDiagramSynthesis.this._kRenderingExtensions.<KRendering>setBackgroundGradient(_kRendering, Colors.WHITE, KlighdConstants.ALPHA_FULL_OPAQUE, Colors.RED, 150, 90);
              }
            };
            _filter_2.forEach(_function_1);
          } else {
            final ArrayList<EClassifier> chosenClasse = Lists.<EClassifier>newArrayList(depictedClasses);
            List<EClassifier> _elvis = null;
            EClassifier _head_1 = null;
            if (depictedClasses!=null) {
              _head_1=IterableExtensions.<EClassifier>head(depictedClasses);
            }
            EPackage _ePackage = null;
            if (_head_1!=null) {
              _ePackage=_head_1.getEPackage();
            }
            EList<EClassifier> _eClassifiers = null;
            if (_ePackage!=null) {
              _eClassifiers=_ePackage.getEClassifiers();
            }
            if (((List<EClassifier>) _eClassifiers) != null) {
              _elvis = ((List<EClassifier>) _eClassifiers);
            } else {
              List<EClassifier> _emptyList = CollectionLiterals.<EClassifier>emptyList();
              _elvis = _emptyList;
            }
            final Procedure1<List<EClassifier>> _function_2 = new Procedure1<List<EClassifier>>() {
              @Override
              public void apply(final List<EClassifier> classes) {
                EcoreDiagramSynthesis.this.createElementFigures(classes, it);
                Iterables.<EClassifier>addAll(depictedClasses, classes);
              }
            };
            ObjectExtensions.<List<EClassifier>>operator_doubleArrow(_elvis, _function_2);
            final Consumer<EClassifier> _function_3 = new Consumer<EClassifier>() {
              @Override
              public void accept(final EClassifier it) {
                KNode _node = EcoreDiagramSynthesis.this._kNodeExtensions.getNode(it);
                KRendering _kRendering = EcoreDiagramSynthesis.this._kRenderingExtensions.getKRendering(_node);
                EcoreDiagramSynthesis.this._kRenderingExtensions.<KRendering>setBackgroundGradient(_kRendering, Colors.WHITE, KlighdConstants.ALPHA_FULL_OPAQUE, Colors.RED, 150, 90);
              }
            };
            chosenClasse.forEach(_function_3);
          }
        }
        Iterable<EPackage> _filter_3 = Iterables.<EPackage>filter(choice, EPackage.class);
        final Consumer<EPackage> _function_4 = new Consumer<EPackage>() {
          @Override
          public void accept(final EPackage ePackage) {
            EList<EClassifier> _eClassifiers = ePackage.getEClassifiers();
            final Procedure1<EList<EClassifier>> _function = new Procedure1<EList<EClassifier>>() {
              @Override
              public void apply(final EList<EClassifier> clazz) {
                EcoreDiagramSynthesis.this.createElementFigures(clazz, it);
              }
            };
            ObjectExtensions.<EList<EClassifier>>operator_doubleArrow(_eClassifiers, _function);
          }
        };
        _filter_3.forEach(_function_4);
      }
    };
    return ObjectExtensions.<KNode>operator_doubleArrow(_createNode, _function);
  }
  
  public void createElementFigures(final Iterable<EClassifier> classes, final KNode rootNode) {
    this.createClassifierFigures(classes, rootNode);
    this.createInheritanceConnections(classes);
    this.createAssociationConnections(classes);
  }
  
  public void createClassifierFigures(final Iterable<EClassifier> classes, final KNode rootNode) {
    Iterable<EClassifier> _filterNull = IterableExtensions.<EClassifier>filterNull(classes);
    final Consumer<EClassifier> _function = new Consumer<EClassifier>() {
      @Override
      public void accept(final EClassifier clazz) {
        EList<KNode> _children = rootNode.getChildren();
        KNode _createNode = EcoreDiagramSynthesis.this._kNodeExtensions.createNode(clazz);
        KNode _associateWith = EcoreDiagramSynthesis.this.<KNode>associateWith(_createNode, clazz);
        final Procedure1<KNode> _function = new Procedure1<KNode>() {
          @Override
          public void apply(final KNode it) {
            KLayoutData _data = it.<KLayoutData>getData(KLayoutData.class);
            KlighdSemanticDiagramData _of = KlighdSemanticDiagramData.of(KlighdConstants.SEMANTIC_DATA_CLASS, "classifier");
            _data.<KlighdSemanticDiagramData>setProperty(KlighdProperties.SEMANTIC_DATA, _of);
            KRectangle _addRectangle = EcoreDiagramSynthesis.this._kRenderingExtensions.addRectangle(it);
            final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
              @Override
              public void apply(final KRectangle it) {
                EcoreDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 1.5f);
                EcoreDiagramSynthesis.this._kRenderingExtensions.setForeground(it, EcoreDiagramSynthesis.LINE_COLOR);
                EcoreDiagramSynthesis.this._kRenderingExtensions.<KRectangle>setBackgroundGradient(it, Colors.LEMON_CHIFFON, 100, Colors.LIGHT_GOLDENROD_1, 255, 90);
                EcoreDiagramSynthesis.this._kRenderingExtensions.setShadow(it, Colors.BLACK);
                KRectangle _addRectangle = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addRectangle(it);
                final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
                  @Override
                  public void apply(final KRectangle it) {
                    EcoreDiagramSynthesis.this._kRenderingExtensions.setInvisible(it, true);
                    KRectangle _addRectangle = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addRectangle(it);
                    final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
                      @Override
                      public void apply(final KRectangle it) {
                        EcoreDiagramSynthesis.this._kRenderingExtensions.setInvisible(it, true);
                        EcoreDiagramSynthesis.this._kRenderingExtensions.<KRectangle>setPointPlacementData(it, EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0.5f, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0.5f, EcoreDiagramSynthesis.this._kRenderingExtensions.H_CENTRAL, EcoreDiagramSynthesis.this._kRenderingExtensions.V_CENTRAL, 0, 0, 0, 0);
                        EClass _eEnum = EcorePackage.eINSTANCE.getEEnum();
                        boolean _isInstance = _eEnum.isInstance(clazz);
                        if (_isInstance) {
                          EcoreDiagramSynthesis.this._kContainerRenderingExtensions.setGridPlacement(it, 1);
                          KText _addText = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addText(it, "<<Enum>>");
                          final Procedure1<KText> _function = new Procedure1<KText>() {
                            @Override
                            public void apply(final KText it) {
                              EcoreDiagramSynthesis.this._kRenderingExtensions.setFontSize(it, 13);
                              EcoreDiagramSynthesis.this._kRenderingExtensions.setFontItalic(it, true);
                              EcoreDiagramSynthesis.this._kRenderingExtensions.setVerticalAlignment(it, EcoreDiagramSynthesis.this._kRenderingExtensions.V_CENTRAL);
                              KGridPlacementData _setGridPlacementData = EcoreDiagramSynthesis.this._kRenderingExtensions.setGridPlacementData(it);
                              KAreaPlacementData _from = EcoreDiagramSynthesis.this._kRenderingExtensions.from(_setGridPlacementData, EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 20, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, 3, 0);
                              EcoreDiagramSynthesis.this._kRenderingExtensions.to(_from, EcoreDiagramSynthesis.this._kRenderingExtensions.RIGHT, 20, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 0, 0f);
                            }
                          };
                          ObjectExtensions.<KText>operator_doubleArrow(_addText, _function);
                          String _name = clazz.getName();
                          String _nullToEmpty = Strings.nullToEmpty(_name);
                          KText _addText_1 = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addText(it, _nullToEmpty);
                          KText _associateWith = EcoreDiagramSynthesis.this.<KText>associateWith(_addText_1, clazz);
                          final Procedure1<KText> _function_1 = new Procedure1<KText>() {
                            @Override
                            public void apply(final KText it) {
                              EcoreDiagramSynthesis.this._kRenderingExtensions.setFontSize(it, 15);
                              EcoreDiagramSynthesis.this._kRenderingExtensions.setFontBold(it, true);
                              KGridPlacementData _setGridPlacementData = EcoreDiagramSynthesis.this._kRenderingExtensions.setGridPlacementData(it);
                              KAreaPlacementData _from = EcoreDiagramSynthesis.this._kRenderingExtensions.from(_setGridPlacementData, EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 20, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, 2, 0f);
                              EcoreDiagramSynthesis.this._kRenderingExtensions.to(_from, EcoreDiagramSynthesis.this._kRenderingExtensions.RIGHT, 20, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 7, 0);
                            }
                          };
                          ObjectExtensions.<KText>operator_doubleArrow(_associateWith, _function_1);
                        } else {
                          String _xifexpression = null;
                          boolean _and = false;
                          if (!(clazz instanceof EClass)) {
                            _and = false;
                          } else {
                            boolean _isInterface = ((EClass) clazz).isInterface();
                            _and = _isInterface;
                          }
                          if (_and) {
                            _xifexpression = "icons/Interface.png";
                          } else {
                            _xifexpression = "icons/Class.png";
                          }
                          KImage _addImage = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addImage(it, "de.cau.cs.kieler.klighd.examples", _xifexpression);
                          KImage _setPointPlacementData = EcoreDiagramSynthesis.this._kRenderingExtensions.<KImage>setPointPlacementData(_addImage, EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 20, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0.5f, EcoreDiagramSynthesis.this._kRenderingExtensions.H_CENTRAL, EcoreDiagramSynthesis.this._kRenderingExtensions.V_CENTRAL, 10, 10, 20, 20);
                          EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addEllipticalClip(_setPointPlacementData);
                          String _name_1 = clazz.getName();
                          String _nullToEmpty_1 = Strings.nullToEmpty(_name_1);
                          KText _addText_2 = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addText(it, _nullToEmpty_1);
                          KText _associateWith_1 = EcoreDiagramSynthesis.this.<KText>associateWith(_addText_2, clazz);
                          final Procedure1<KText> _function_2 = new Procedure1<KText>() {
                            @Override
                            public void apply(final KText it) {
                              KlighdSemanticDiagramData _of = KlighdSemanticDiagramData.of(KlighdConstants.SEMANTIC_DATA_CLASS, "classifierText");
                              it.<KlighdSemanticDiagramData>setProperty(KlighdProperties.SEMANTIC_DATA, _of);
                              EcoreDiagramSynthesis.this._kRenderingExtensions.setFontSize(it, 15);
                              EcoreDiagramSynthesis.this._kRenderingExtensions.setFontBold(it, true);
                              boolean _and = false;
                              if (!(clazz instanceof EClass)) {
                                _and = false;
                              } else {
                                boolean _isAbstract = ((EClass) clazz).isAbstract();
                                _and = _isAbstract;
                              }
                              EcoreDiagramSynthesis.this._kRenderingExtensions.setFontItalic(it, _and);
                              EcoreDiagramSynthesis.this._kRenderingExtensions.<KText>setPointPlacementData(it, EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 40, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0.5f, EcoreDiagramSynthesis.this._kRenderingExtensions.H_LEFT, EcoreDiagramSynthesis.this._kRenderingExtensions.V_CENTRAL, 10, 10, 0, 0);
                            }
                          };
                          ObjectExtensions.<KText>operator_doubleArrow(_associateWith_1, _function_2);
                        }
                      }
                    };
                    ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function);
                  }
                };
                ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function);
                boolean _booleanValue = EcoreDiagramSynthesis.this.getBooleanValue(EcoreDiagramSynthesis.ATTRIBUTES);
                boolean _not = (!_booleanValue);
                if (_not) {
                  return;
                }
                KGridPlacement _setGridPlacement = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.setGridPlacement(it, 1);
                KGridPlacement _from = EcoreDiagramSynthesis.this._kRenderingExtensions.from(_setGridPlacement, EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 2, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, 2, 0);
                EcoreDiagramSynthesis.this._kRenderingExtensions.to(_from, EcoreDiagramSynthesis.this._kRenderingExtensions.RIGHT, 2, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 2, 0);
                boolean _and = false;
                EClass _eClass = EcorePackage.eINSTANCE.getEClass();
                boolean _isInstance = _eClass.isInstance(clazz);
                if (!_isInstance) {
                  _and = false;
                } else {
                  EList<EAttribute> _eAttributes = ((EClass) clazz).getEAttributes();
                  boolean _isEmpty = _eAttributes.isEmpty();
                  boolean _not_1 = (!_isEmpty);
                  _and = _not_1;
                }
                if (_and) {
                  KPolyline _addHorizontalLine = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addHorizontalLine(it, 1, 1.5f);
                  EcoreDiagramSynthesis.this._kRenderingExtensions.addStyleRef(_addHorizontalLine, it);
                  KRectangle _addRectangle_1 = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addRectangle(it);
                  final Procedure1<KRectangle> _function_1 = new Procedure1<KRectangle>() {
                    @Override
                    public void apply(final KRectangle it) {
                      EcoreDiagramSynthesis.this._kRenderingExtensions.setInvisible(it, true);
                      EcoreDiagramSynthesis.this._kRenderingExtensions.<KRectangle>setSurroundingSpaceGrid(it, 5, 0);
                      KGridPlacement _setGridPlacement = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.setGridPlacement(it, 1);
                      EcoreDiagramSynthesis.this._kRenderingExtensions.from(_setGridPlacement, EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, (-2), 0);
                      EList<EAttribute> _eAttributes = ((EClass) clazz).getEAttributes();
                      final Consumer<EAttribute> _function = new Consumer<EAttribute>() {
                        @Override
                        public void accept(final EAttribute attr) {
                          KRectangle _addRectangle = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addRectangle(it);
                          final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
                            @Override
                            public void apply(final KRectangle it) {
                              EcoreDiagramSynthesis.this._kRenderingExtensions.setInvisible(it, true);
                              KRectangle _addAttributeIcon = EcoreDiagramSynthesis.this.addAttributeIcon(it);
                              EcoreDiagramSynthesis.this._kRenderingExtensions.<KRectangle>setPointPlacementData(_addAttributeIcon, EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 10, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, 1.5f, 0.5f, EcoreDiagramSynthesis.this._kRenderingExtensions.H_CENTRAL, EcoreDiagramSynthesis.this._kRenderingExtensions.V_CENTRAL, 0, 0, 15f, 7.5f);
                              String _xifexpression = null;
                              EDataType _eAttributeType = attr.getEAttributeType();
                              if ((_eAttributeType instanceof EEnum)) {
                                EDataType _eAttributeType_1 = attr.getEAttributeType();
                                _xifexpression = _eAttributeType_1.getName();
                              } else {
                                EDataType _eAttributeType_2 = attr.getEAttributeType();
                                _xifexpression = _eAttributeType_2.getInstanceTypeName();
                              }
                              final String typeName = _xifexpression;
                              String _name = attr.getName();
                              String _plus = (_name + " : ");
                              String _replaceFirst = typeName.replaceFirst("java.lang.", "");
                              String _plus_1 = (_plus + _replaceFirst);
                              KText _addText = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addText(it, _plus_1);
                              final Procedure1<KText> _function = new Procedure1<KText>() {
                                @Override
                                public void apply(final KText it) {
                                  EcoreDiagramSynthesis.this._kRenderingExtensions.setFontSize(it, 13);
                                  EcoreDiagramSynthesis.this._kRenderingExtensions.setHorizontalAlignment(it, EcoreDiagramSynthesis.this._kRenderingExtensions.H_LEFT);
                                  EcoreDiagramSynthesis.this._kRenderingExtensions.setVerticalAlignment(it, EcoreDiagramSynthesis.this._kRenderingExtensions.V_CENTRAL);
                                  EcoreDiagramSynthesis.this._kRenderingExtensions.<KText>setPointPlacementData(it, EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 25, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0.5f, EcoreDiagramSynthesis.this._kRenderingExtensions.H_LEFT, EcoreDiagramSynthesis.this._kRenderingExtensions.V_CENTRAL, 10, 3, 0, 0);
                                }
                              };
                              ObjectExtensions.<KText>operator_doubleArrow(_addText, _function);
                            }
                          };
                          ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function);
                        }
                      };
                      _eAttributes.forEach(_function);
                    }
                  };
                  ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle_1, _function_1);
                }
                EClass _eEnum = EcorePackage.eINSTANCE.getEEnum();
                boolean _isInstance_1 = _eEnum.isInstance(clazz);
                if (_isInstance_1) {
                  EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addHorizontalLine(it, 1, 1.5f);
                  KRectangle _addRectangle_2 = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addRectangle(it);
                  final Procedure1<KRectangle> _function_2 = new Procedure1<KRectangle>() {
                    @Override
                    public void apply(final KRectangle rect) {
                      EcoreDiagramSynthesis.this._kRenderingExtensions.setInvisible(rect, true);
                      EcoreDiagramSynthesis.this._kRenderingExtensions.setForeground(rect, Colors.RED);
                      EcoreDiagramSynthesis.this._kRenderingExtensions.<KRectangle>setSurroundingSpaceGrid(rect, 5, 0);
                      KGridPlacement _setGridPlacement = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.setGridPlacement(rect, 1);
                      EcoreDiagramSynthesis.this._kRenderingExtensions.to(_setGridPlacement, EcoreDiagramSynthesis.this._kRenderingExtensions.RIGHT, 0, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 0, 0);
                      EList<EEnumLiteral> _eLiterals = ((EEnum) clazz).getELiterals();
                      final Consumer<EEnumLiteral> _function = new Consumer<EEnumLiteral>() {
                        @Override
                        public void accept(final EEnumLiteral it) {
                          String _name = it.getName();
                          KText _addText = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addText(rect, _name);
                          final Procedure1<KText> _function = new Procedure1<KText>() {
                            @Override
                            public void apply(final KText it) {
                              EcoreDiagramSynthesis.this._kRenderingExtensions.setHorizontalAlignment(it, EcoreDiagramSynthesis.this._kRenderingExtensions.H_CENTRAL);
                              EcoreDiagramSynthesis.this._kRenderingExtensions.setVerticalAlignment(it, EcoreDiagramSynthesis.this._kRenderingExtensions.V_CENTRAL);
                              EcoreDiagramSynthesis.this._kRenderingExtensions.<KText>setSurroundingSpaceGrid(it, 3, 0);
                            }
                          };
                          ObjectExtensions.<KText>operator_doubleArrow(_addText, _function);
                        }
                      };
                      _eLiterals.forEach(_function);
                    }
                  };
                  ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle_2, _function_2);
                }
              }
            };
            ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function);
          }
        };
        KNode _doubleArrow = ObjectExtensions.<KNode>operator_doubleArrow(_associateWith, _function);
        _children.add(_doubleArrow);
      }
    };
    _filterNull.forEach(_function);
  }
  
  private Set<EReference> visitedRefs = CollectionLiterals.<EReference>newHashSet();
  
  public void createAssociationConnections(final Iterable<EClassifier> classes) {
    final List<EClassifier> list = IterableExtensions.<EClassifier>toList(classes);
    Iterable<EClass> _filter = Iterables.<EClass>filter(list, EClass.class);
    final Consumer<EClass> _function = new Consumer<EClass>() {
      @Override
      public void accept(final EClass it) {
        EList<EStructuralFeature> _eStructuralFeatures = it.getEStructuralFeatures();
        Iterable<EReference> _filter = Iterables.<EReference>filter(_eStructuralFeatures, EReference.class);
        final Function1<EReference, Boolean> _function = new Function1<EReference, Boolean>() {
          @Override
          public Boolean apply(final EReference it) {
            EClassifier _eType = it.getEType();
            return Boolean.valueOf(list.contains(_eType));
          }
        };
        Iterable<EReference> _filter_1 = IterableExtensions.<EReference>filter(_filter, _function);
        final Function1<EReference, Boolean> _function_1 = new Function1<EReference, Boolean>() {
          @Override
          public Boolean apply(final EReference it) {
            boolean _xblockexpression = false;
            {
              final EReference opposite = it.getEOpposite();
              boolean _or = false;
              boolean _or_1 = false;
              boolean _isContainment = it.isContainment();
              if (_isContainment) {
                _or_1 = true;
              } else {
                boolean _equals = Objects.equal(opposite, null);
                _or_1 = _equals;
              }
              if (_or_1) {
                _or = true;
              } else {
                boolean _and = false;
                boolean _isContainment_1 = opposite.isContainment();
                boolean _not = (!_isContainment_1);
                if (!_not) {
                  _and = false;
                } else {
                  boolean _contains = EcoreDiagramSynthesis.this.visitedRefs.contains(opposite);
                  boolean _not_1 = (!_contains);
                  _and = _not_1;
                }
                _or = _and;
              }
              _xblockexpression = _or;
            }
            return Boolean.valueOf(_xblockexpression);
          }
        };
        Iterable<EReference> _filter_2 = IterableExtensions.<EReference>filter(_filter_1, _function_1);
        final Consumer<EReference> _function_2 = new Consumer<EReference>() {
          @Override
          public void accept(final EReference it) {
            EcoreDiagramSynthesis.this.createAssociationConnection(it);
          }
        };
        _filter_2.forEach(_function_2);
      }
    };
    _filter.forEach(_function);
  }
  
  public KEdge createAssociationConnection(final EReference ref) {
    KEdge _xblockexpression = null;
    {
      this.visitedRefs.add(ref);
      KEdge _createEdge = this._kEdgeExtensions.createEdge(ref);
      final Procedure1<KEdge> _function = new Procedure1<KEdge>() {
        @Override
        public void apply(final KEdge it) {
          EcoreDiagramSynthesis.this._kEdgeExtensions.<EdgeType>addLayoutParam(it, CoreOptions.EDGE_TYPE, EdgeType.ASSOCIATION);
          EObject _eContainer = ref.eContainer();
          KNode _node = EcoreDiagramSynthesis.this._kNodeExtensions.getNode(_eContainer);
          it.setSource(_node);
          EClassifier _eType = ref.getEType();
          KNode _node_1 = EcoreDiagramSynthesis.this._kNodeExtensions.getNode(_eType);
          it.setTarget(_node_1);
          KLayoutData _data = it.<KLayoutData>getData(KLayoutData.class);
          KlighdSemanticDiagramData _of = KlighdSemanticDiagramData.of(KlighdConstants.SEMANTIC_DATA_CLASS, "association");
          _data.<KlighdSemanticDiagramData>setProperty(KlighdProperties.SEMANTIC_DATA, _of);
          KPolyline _addPolyline = EcoreDiagramSynthesis.this._kEdgeExtensions.addPolyline(it);
          final Procedure1<KPolyline> _function = new Procedure1<KPolyline>() {
            @Override
            public void apply(final KPolyline it) {
              EcoreDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 2);
              EcoreDiagramSynthesis.this._kRenderingExtensions.setForeground(it, EcoreDiagramSynthesis.LINE_COLOR);
              EReference _eOpposite = ref.getEOpposite();
              boolean _equals = Objects.equal(_eOpposite, null);
              if (_equals) {
                EcoreDiagramSynthesis.this._kPolylineExtensions.addHeadArrowDecorator(it);
              }
              boolean _isContainment = ref.isContainment();
              if (_isContainment) {
                KPolygon _addPolygon = EcoreDiagramSynthesis.this._kContainerRenderingExtensions.addPolygon(it);
                final Procedure1<KPolygon> _function = new Procedure1<KPolygon>() {
                  @Override
                  public void apply(final KPolygon it) {
                    EList<KPosition> _points = it.getPoints();
                    KPosition _createKPosition = EcoreDiagramSynthesis.this._kRenderingExtensions.createKPosition(EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0.5f);
                    _points.add(_createKPosition);
                    EList<KPosition> _points_1 = it.getPoints();
                    KPosition _createKPosition_1 = EcoreDiagramSynthesis.this._kRenderingExtensions.createKPosition(EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0.5f, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0);
                    _points_1.add(_createKPosition_1);
                    EList<KPosition> _points_2 = it.getPoints();
                    KPosition _createKPosition_2 = EcoreDiagramSynthesis.this._kRenderingExtensions.createKPosition(EcoreDiagramSynthesis.this._kRenderingExtensions.RIGHT, 0, 0, EcoreDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0.5f);
                    _points_2.add(_createKPosition_2);
                    EList<KPosition> _points_3 = it.getPoints();
                    KPosition _createKPosition_3 = EcoreDiagramSynthesis.this._kRenderingExtensions.createKPosition(EcoreDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0.5f, EcoreDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 0, 0);
                    _points_3.add(_createKPosition_3);
                    EcoreDiagramSynthesis.this._kRenderingExtensions.<KPolygon>setDecoratorPlacementData(it, 24, 12, 12, 0, true);
                    EcoreDiagramSynthesis.this._kRenderingExtensions.setForeground(it, EcoreDiagramSynthesis.LINE_COLOR);
                    EcoreDiagramSynthesis.this._kRenderingExtensions.setBackground(it, EcoreDiagramSynthesis.LINE_COLOR);
                  }
                };
                ObjectExtensions.<KPolygon>operator_doubleArrow(_addPolygon, _function);
              }
            }
          };
          ObjectExtensions.<KPolyline>operator_doubleArrow(_addPolyline, _function);
        }
      };
      _xblockexpression = ObjectExtensions.<KEdge>operator_doubleArrow(_createEdge, _function);
    }
    return _xblockexpression;
  }
  
  public void createInheritanceConnections(final Iterable<EClassifier> classes) {
    final List<EClassifier> list = IterableExtensions.<EClassifier>toList(classes);
    Iterable<EClass> _filter = Iterables.<EClass>filter(list, EClass.class);
    final Consumer<EClass> _function = new Consumer<EClass>() {
      @Override
      public void accept(final EClass child) {
        EList<EClass> _eSuperTypes = child.getESuperTypes();
        final Function1<EClass, Boolean> _function = new Function1<EClass, Boolean>() {
          @Override
          public Boolean apply(final EClass it) {
            return Boolean.valueOf(list.contains(it));
          }
        };
        Iterable<EClass> _filter = IterableExtensions.<EClass>filter(_eSuperTypes, _function);
        final Consumer<EClass> _function_1 = new Consumer<EClass>() {
          @Override
          public void accept(final EClass parent) {
            EcoreDiagramSynthesis.this.createInheritanceConnection(child, parent);
          }
        };
        _filter.forEach(_function_1);
      }
    };
    _filter.forEach(_function);
  }
  
  public KEdge createInheritanceConnection(final EClass child, final EClass parent) {
    org.eclipse.xtext.xbase.lib.Pair<EClass, EClass> _pair = new org.eclipse.xtext.xbase.lib.Pair<EClass, EClass>(child, parent);
    KEdge _createEdge = this._kEdgeExtensions.createEdge(_pair);
    final Procedure1<KEdge> _function = new Procedure1<KEdge>() {
      @Override
      public void apply(final KEdge it) {
        EcoreDiagramSynthesis.this._kEdgeExtensions.<EdgeType>addLayoutParam(it, CoreOptions.EDGE_TYPE, EdgeType.GENERALIZATION);
        KLayoutData _data = it.<KLayoutData>getData(KLayoutData.class);
        KlighdSemanticDiagramData _of = KlighdSemanticDiagramData.of(KlighdConstants.SEMANTIC_DATA_CLASS, "inheritence");
        _data.<KlighdSemanticDiagramData>setProperty(KlighdProperties.SEMANTIC_DATA, _of);
        KNode _node = EcoreDiagramSynthesis.this._kNodeExtensions.getNode(child);
        it.setSource(_node);
        KNode _node_1 = EcoreDiagramSynthesis.this._kNodeExtensions.getNode(parent);
        it.setTarget(_node_1);
        it.getData();
        KPolyline _addPolyline = EcoreDiagramSynthesis.this._kEdgeExtensions.addPolyline(it);
        final Procedure1<KPolyline> _function = new Procedure1<KPolyline>() {
          @Override
          public void apply(final KPolyline it) {
            EcoreDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 2);
            EcoreDiagramSynthesis.this._kRenderingExtensions.setForeground(it, EcoreDiagramSynthesis.LINE_COLOR);
            EcoreDiagramSynthesis.this._kPolylineExtensions.addInheritanceTriangleArrowDecorator(it);
          }
        };
        ObjectExtensions.<KPolyline>operator_doubleArrow(_addPolyline, _function);
      }
    };
    return ObjectExtensions.<KEdge>operator_doubleArrow(_createEdge, _function);
  }
  
  public KRectangle addAttributeIcon(final KContainerRendering parent) {
    KRectangle _addRectangle = this._kContainerRenderingExtensions.addRectangle(parent);
    final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
      @Override
      public void apply(final KRectangle it) {
        EcoreDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 1.75f);
        EcoreDiagramSynthesis.this._kRenderingExtensions.<KRectangle>setForegroundGradient(it, Colors.GOLDENROD_4, 255, Colors.DIM_GRAY, 255, 90);
        EcoreDiagramSynthesis.this._kRenderingExtensions.setBackground(it, Colors.LEMON_CHIFFON);
      }
    };
    return ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function);
  }
}
