/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.klighd.krendering.extensions;

import com.google.inject.Injector;
import com.google.inject.Scope;
import de.cau.cs.kieler.klighd.krendering.ViewSynthesisShared;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashMap;
import javax.inject.Inject;
import org.eclipse.elk.core.klayoutdata.KLayoutData;
import org.eclipse.elk.core.klayoutdata.KShapeLayout;
import org.eclipse.elk.core.math.KVector;
import org.eclipse.elk.core.options.CoreOptions;
import org.eclipse.elk.core.options.NodeLabelPlacement;
import org.eclipse.elk.core.util.ElkUtil;
import org.eclipse.elk.core.util.Pair;
import org.eclipse.elk.graph.KGraphElement;
import org.eclipse.elk.graph.KNode;
import org.eclipse.elk.graph.properties.IProperty;
import org.eclipse.elk.graph.properties.Property;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * Provides some helpful extension methods for simplifying the composition of KGraph/KRendering-based view models.<br>
 * <br>
 * In order to employ them beyond KLighD diagram syntheses you best declare a field of type
 * {@link KNodeExtensions} in your class and annotate it with {@link Inject Inject}.<br>
 * <br>
 * Make sure to bind the {@link ViewSynthesisShared} annotation in the employed
 * {@link Injector Injector} to a {@link Scope}, e.g. by calling
 * {@code Guice.createInjector(KRenderingExtensionsPlugin.createSingletonScopeBindingModule());} or
 * {@code Guice.createInjector(KRenderingExtensionsPlugin.createNoScopeBindingModule());}.<br>
 * <br>
 * By means of that {@link Injector Injector} you may get a new instance of your class,
 * or you may inject the above mentioned attribute within instances of your class, e.g. by calling
 * {@code injector.injectMembers(this)} in the constructor of your class.
 * 
 * @author chsch
 * 
 * @containsExtensions
 */
@ViewSynthesisShared
@SuppressWarnings("all")
public class KNodeExtensions {
  /**
   * A convenient getter preserving the element image relation by a create extension.
   */
  private KNode internalCreateNode(final ArrayList<Object> oc) {
    final ArrayList<?> _cacheKey = CollectionLiterals.newArrayList(oc);
    final KNode _result;
    synchronized (_createCache_internalCreateNode) {
      if (_createCache_internalCreateNode.containsKey(_cacheKey)) {
        return _createCache_internalCreateNode.get(_cacheKey);
      }
      KNode _createInitializedNode = ElkUtil.createInitializedNode();
      _result = _createInitializedNode;
      _createCache_internalCreateNode.put(_cacheKey, _result);
    }
    _init_internalCreateNode(_result, oc);
    return _result;
  }
  
  private final HashMap<ArrayList<?>, KNode> _createCache_internalCreateNode = CollectionLiterals.newHashMap();
  
  private void _init_internalCreateNode(final KNode node, final ArrayList<Object> oc) {
  }
  
  /**
   * A convenient getter preserving the element image relation by a create extension.
   */
  public KNode getNode(final Object o) {
    ArrayList<Object> _newArrayList = CollectionLiterals.<Object>newArrayList(o);
    return this.internalCreateNode(_newArrayList);
  }
  
  /**
   * A convenient getter preserving the element image relation by a create extension.
   */
  public KNode getNode(final Object o1, final Object o2) {
    ArrayList<Object> _newArrayList = CollectionLiterals.<Object>newArrayList(o1, o2);
    return this.internalCreateNode(_newArrayList);
  }
  
  /**
   * A convenience method to create a KNode without relating it to a business object.
   */
  public KNode createNode() {
    return ElkUtil.createInitializedNode();
  }
  
  /**
   * An alias of {@link #getNode} allowing to express in business that the KNode will
   * be created at this place. It is just syntactic sugar.
   */
  public KNode createNode(final Object o) {
    return this.getNode(o);
  }
  
  /**
   * An alias of {@link #getNode} allowing to express in business that the KNode will
   * be created at this place. It is just syntactic sugar.
   */
  public KNode createNode(final Object o1, final Object o2) {
    return this.getNode(o1, o2);
  }
  
  public Pair<Float, Float> getNodeSize(final KNode node) {
    Pair<Float, Float> _pair = new Pair<Float, Float>();
    final Procedure1<Pair<Float, Float>> _function = new Procedure1<Pair<Float, Float>>() {
      @Override
      public void apply(final Pair<Float, Float> it) {
        final KShapeLayout layout = node.<KShapeLayout>getData(KShapeLayout.class);
        float _height = layout.getHeight();
        it.setFirst(Float.valueOf(_height));
        float _height_1 = layout.getHeight();
        it.setSecond(Float.valueOf(_height_1));
      }
    };
    return ObjectExtensions.<Pair<Float, Float>>operator_doubleArrow(_pair, _function);
  }
  
  public float getHeight(final KNode node) {
    KShapeLayout _data = node.<KShapeLayout>getData(KShapeLayout.class);
    return _data.getHeight();
  }
  
  /**
   * Is used in KPortExtensions
   */
  public float getWidth(final KNode node) {
    KShapeLayout _data = node.<KShapeLayout>getData(KShapeLayout.class);
    return _data.getWidth();
  }
  
  public KNode setNodeSize(final KNode node, final float width, final float height) {
    final Procedure1<KNode> _function = new Procedure1<KNode>() {
      @Override
      public void apply(final KNode it) {
        KShapeLayout _data = it.<KShapeLayout>getData(KShapeLayout.class);
        _data.setSize(width, height);
        KNodeExtensions.this.setMinimalNodeSize(it, width, height);
      }
    };
    return ObjectExtensions.<KNode>operator_doubleArrow(node, _function);
  }
  
  public KNode setWidth(final KNode node, final float width) {
    KShapeLayout _data = node.<KShapeLayout>getData(KShapeLayout.class);
    final float height = _data.getHeight();
    this.setNodeSize(node, width, height);
    return node;
  }
  
  public KNode setHeight(final KNode node, final float height) {
    KShapeLayout _data = node.<KShapeLayout>getData(KShapeLayout.class);
    final float width = _data.getWidth();
    this.setNodeSize(node, width, height);
    return node;
  }
  
  private final static IProperty<KVector> MINIMAL_NODE_SIZE = new Property<KVector>(
    "de.cau.cs.kieler.klighd.minimalNodeSize", new KVector(10d, 10d));
  
  public KNode setMinimalNodeSize(final KNode node, final float width, final float height) {
    KVector _kVector = new KVector(width, height);
    return this.<KVector>addLayoutParam(node, KNodeExtensions.MINIMAL_NODE_SIZE, _kVector);
  }
  
  public <T extends Object> KNode addLayoutParam(final KNode node, final IProperty<? super T> property, final T value) {
    KShapeLayout _data = null;
    if (node!=null) {
      _data=node.<KShapeLayout>getData(KShapeLayout.class);
    }
    if (_data!=null) {
      _data.<T>setProperty(property, value);
    }
    return node;
  }
  
  /**
   * Internal helper for setting layout options without the need to check for KNode, KEdge, ...
   */
  private <S extends Object, T extends KGraphElement> T setLayoutOption(final T kgraphElement, final IProperty<S> option, final S value) {
    KLayoutData _data = null;
    if (kgraphElement!=null) {
      _data=kgraphElement.<KLayoutData>getData(KLayoutData.class);
    }
    if (_data!=null) {
      _data.<S>setProperty(option, value);
    }
    return kgraphElement;
  }
  
  /**
   * Configures labels on this node to be outside bottom left-aligned labels by default
   */
  public KNode configureOutsideBottomLeftNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _outsideBottomLeft = NodeLabelPlacement.outsideBottomLeft();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _outsideBottomLeft);
  }
  
  /**
   * Configures labels on this node to be outside bottom centrally-aligned labels by default
   */
  public KNode configureOutsideBottomCenteredNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _outsideBottomCenter = NodeLabelPlacement.outsideBottomCenter();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _outsideBottomCenter);
  }
  
  /**
   * Configures labels on this node to be outside bottom right-aligned labels by default
   */
  public KNode configureOutsideBottomRightNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _outsideBottomRight = NodeLabelPlacement.outsideBottomRight();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _outsideBottomRight);
  }
  
  /**
   * Configures labels on this node to be outside top left-aligned labels by default
   */
  public KNode configureOutsideTopLeftNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _outsideTopLeft = NodeLabelPlacement.outsideTopLeft();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _outsideTopLeft);
  }
  
  /**
   * Configures labels on this node to be outside top centrally-aligned labels by default
   */
  public KNode configureOutsideTopCenteredNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _outsideTopCenter = NodeLabelPlacement.outsideTopCenter();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _outsideTopCenter);
  }
  
  /**
   * Configures labels on this node to be outside top right-aligned labels by default
   */
  public KNode configureOutsideTopRightNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _outsideTopRight = NodeLabelPlacement.outsideTopRight();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _outsideTopRight);
  }
  
  /**
   * Configures labels on this node to be inside bottom left-aligned labels by default
   */
  public KNode configureInsideBottomLeftNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _insideBottomLeft = NodeLabelPlacement.insideBottomLeft();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _insideBottomLeft);
  }
  
  /**
   * Configures labels on this node to be inside bottom centrally-aligned labels by default
   */
  public KNode configureInsideBottomCenteredNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _insideBottomCenter = NodeLabelPlacement.insideBottomCenter();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _insideBottomCenter);
  }
  
  /**
   * Configures labels on this node to be inside bottom right-aligned labels by default
   */
  public KNode configureInsideBottomRightNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _insideBottomRight = NodeLabelPlacement.insideBottomRight();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _insideBottomRight);
  }
  
  /**
   * Configures labels on this node to be inside centrally-aligned labels by default
   */
  public KNode configureInsideCenteredNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _insideCenter = NodeLabelPlacement.insideCenter();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _insideCenter);
  }
  
  /**
   * Configures labels on this node to be inside top left-aligned labels by default
   */
  public KNode configureInsideTopLeftNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _insideTopLeft = NodeLabelPlacement.insideTopLeft();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _insideTopLeft);
  }
  
  /**
   * Configures labels on this node to be inside top centrally-aligned labels by default
   */
  public KNode configureInsideTopCenteredNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _insideTopCenter = NodeLabelPlacement.insideTopCenter();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _insideTopCenter);
  }
  
  /**
   * Configures labels on this node to be inside top right-aligned labels by default
   */
  public KNode configureInsideTopRightNodeLabelPlacement(final KNode node) {
    EnumSet<NodeLabelPlacement> _insideTopRight = NodeLabelPlacement.insideTopRight();
    return this.<EnumSet<NodeLabelPlacement>, KNode>setLayoutOption(node, CoreOptions.NODE_LABELS_PLACEMENT, _insideTopRight);
  }
}
