/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2013 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.klighd.xtext.transformations;

import com.google.common.base.Objects;
import com.google.common.base.Predicate;
import com.google.common.collect.ImmutableList;
import com.google.inject.Inject;
import de.cau.cs.kieler.klighd.KlighdConstants;
import de.cau.cs.kieler.klighd.SynthesisOption;
import de.cau.cs.kieler.klighd.ViewContext;
import de.cau.cs.kieler.klighd.actions.FocusAndContextAction;
import de.cau.cs.kieler.klighd.krendering.HorizontalAlignment;
import de.cau.cs.kieler.klighd.krendering.KColor;
import de.cau.cs.kieler.klighd.krendering.KEllipse;
import de.cau.cs.kieler.klighd.krendering.KPolyline;
import de.cau.cs.kieler.klighd.krendering.KRectangle;
import de.cau.cs.kieler.klighd.krendering.KRendering;
import de.cau.cs.kieler.klighd.krendering.KRenderingFactory;
import de.cau.cs.kieler.klighd.krendering.KRenderingLibrary;
import de.cau.cs.kieler.klighd.krendering.KRenderingRef;
import de.cau.cs.kieler.klighd.krendering.KRoundedRectangle;
import de.cau.cs.kieler.klighd.krendering.KSpline;
import de.cau.cs.kieler.klighd.krendering.KStyleHolder;
import de.cau.cs.kieler.klighd.krendering.KText;
import de.cau.cs.kieler.klighd.krendering.VerticalAlignment;
import de.cau.cs.kieler.klighd.krendering.extensions.KColorExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KLibraryExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KPolylineExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KRenderingExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.PositionReferenceX;
import de.cau.cs.kieler.klighd.krendering.extensions.PositionReferenceY;
import de.cau.cs.kieler.klighd.labels.AbstractKlighdLabelManager;
import de.cau.cs.kieler.klighd.labels.ConditionLabelManager;
import de.cau.cs.kieler.klighd.labels.IdentLabelManager;
import de.cau.cs.kieler.klighd.labels.LabelPredicates;
import de.cau.cs.kieler.klighd.labels.SoftWrappingLabelManager;
import de.cau.cs.kieler.klighd.labels.TruncatingLabelManager;
import de.cau.cs.kieler.klighd.syntheses.AbstractDiagramSynthesis;
import de.cau.cs.kieler.klighd.util.KlighdProperties;
import de.cau.cs.kieler.klighd.util.KlighdSynthesisProperties;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Consumer;
import org.eclipse.elk.core.klayoutdata.KLayoutData;
import org.eclipse.elk.core.labels.ILabelManager;
import org.eclipse.elk.core.labels.LabelManagementOptions;
import org.eclipse.elk.core.options.CoreOptions;
import org.eclipse.elk.core.options.EdgeLabelPlacement;
import org.eclipse.elk.core.options.EdgeRouting;
import org.eclipse.elk.core.util.ElkUtil;
import org.eclipse.elk.core.util.GraphDataUtil;
import org.eclipse.elk.core.util.Pair;
import org.eclipse.elk.graph.EMapPropertyHolder;
import org.eclipse.elk.graph.KEdge;
import org.eclipse.elk.graph.KGraphData;
import org.eclipse.elk.graph.KGraphElement;
import org.eclipse.elk.graph.KLabel;
import org.eclipse.elk.graph.KNode;
import org.eclipse.elk.graph.KPort;
import org.eclipse.elk.graph.properties.IProperty;
import org.eclipse.elk.graph.properties.Property;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * Synthesizes a copy of the given {@code KNode} and adds default stuff.
 * 
 * <p>Default stuff, in this case, is the following:</p>
 * <ul>
 *   <li>If a node has no label, but an ID, a label with the ID as its text is added.</li>
 *   <li>If a node label is present, node label placement defaults to inner centered placement.</li>
 *   <li>Edges without attached {@code KRendering}s are provided with a {@code KPolyline} with
 *       an arrowhead at the end. (the default would be a simple polyline without arrowhead)</li>
 *   <li>Labels without attached {@code KRendering}s are provided with a default {@code KText}
 *       rendering that ensures that label sizes will be calculated correctly.</li>
 *   <li>Edge labels without edge label placement get a shiny new {@link EdgeLabelPlacement#CENTER}
 *       placement.</li>
 * </ul>
 * 
 * @author cds
 * @author uru
 */
@SuppressWarnings("all")
public class KGraphDiagramSynthesis extends AbstractDiagramSynthesis<KNode> {
  /**
   * ID this synthesis is registered with KLighD with.
   */
  public final static String TRANSFORMATION_ID = "de.cau.cs.kieler.klighd.xtext.transformations.KGraphDiagramSynthesis";
  
  private final static Predicate<EMapPropertyHolder> PREDICATE_IS_KGRAPHDATA = new Predicate<EMapPropertyHolder>() {
    @Override
    public boolean apply(final EMapPropertyHolder input) {
      return (input instanceof KGraphData);
    }
  };
  
  private final static ImmutableList<IProperty<?>> KNOWN_PROPS = ImmutableList.<IProperty<? extends Object>>of(KlighdProperties.EXPAND, 
    KlighdProperties.EXPANDED_RENDERING, KlighdProperties.COLLAPSED_RENDERING, 
    KlighdProperties.TOOLTIP, KlighdProperties.NOT_SELECTABLE);
  
  /**
   * Whether the model wants default defaults. This property is no layout option, not registered with
   * KIML, and has to be loaded explicitly by the
   * {@link de.cau.cs.kieler.core.kgraph.text.KGraphResource KGraphResource}.
   */
  private final static IProperty<Boolean> DEFAULTS_PROPERTY = new Property<Boolean>(
    "de.cau.cs.kieler.kgraphsynthesis.defaults", Boolean.valueOf(false));
  
  private final static String DEFAULTS_AS_IN_MODEL = "As in Model";
  
  private final static String DEFAULTS_ON = "On";
  
  private final static String DEFAULTS_OFF = "Off";
  
  /**
   * Synthesis option specifying whether default values should be used. Default values are, eg, node
   * size if not specified and port ids as labels if no labels exist.
   */
  private final static SynthesisOption DEFAULTS = SynthesisOption.createChoiceOption("Default Values", 
    ImmutableList.<String>of(KGraphDiagramSynthesis.DEFAULTS_AS_IN_MODEL, KGraphDiagramSynthesis.DEFAULTS_ON, KGraphDiagramSynthesis.DEFAULTS_OFF), KGraphDiagramSynthesis.DEFAULTS_AS_IN_MODEL);
  
  private final static String STYLE_BORING = "Boring";
  
  private final static String STYLE_STYLISH = "Stylish";
  
  private final static String STYLE_HELLO_KITTY = "Hello Kitty";
  
  private final static String STYLE_SPONGEBOB = "Spongebob";
  
  /**
   * The style to be used when drawing the KGraphs.
   */
  private final static SynthesisOption STYLE = SynthesisOption.createChoiceOption(
    "Style", 
    ImmutableList.<String>of(KGraphDiagramSynthesis.STYLE_BORING, KGraphDiagramSynthesis.STYLE_STYLISH, KGraphDiagramSynthesis.STYLE_HELLO_KITTY, KGraphDiagramSynthesis.STYLE_SPONGEBOB), 
    KGraphDiagramSynthesis.STYLE_STYLISH);
  
  private final static String LABELS_NO_LABELS = "No Labels";
  
  private final static String LABELS_TRUNCATE = "Truncate";
  
  private final static String LABELS_SOFT_WORD_WRAP = "Soft Word Wrapping";
  
  private final static String LABELS_FULL = "Full Labels";
  
  private final static SynthesisOption LABEL_SHORTENING_STRATEGY = SynthesisOption.createChoiceOption(
    "Label Shortening Strategy", 
    ImmutableList.<String>of(KGraphDiagramSynthesis.LABELS_NO_LABELS, KGraphDiagramSynthesis.LABELS_TRUNCATE, KGraphDiagramSynthesis.LABELS_SOFT_WORD_WRAP, KGraphDiagramSynthesis.LABELS_FULL), 
    KGraphDiagramSynthesis.LABELS_FULL);
  
  /**
   * {@inheritDoc}
   */
  @Override
  public List<Pair<IProperty<?>, List<?>>> getDisplayedLayoutOptions() {
    return ImmutableList.<Pair<IProperty<?>, List<?>>>of();
  }
  
  /**
   * {@inheritDoc}
   */
  @Override
  public List<SynthesisOption> getDisplayedSynthesisOptions() {
    return ImmutableList.<SynthesisOption>of(
      KGraphDiagramSynthesis.DEFAULTS, 
      KGraphDiagramSynthesis.STYLE, 
      KGraphDiagramSynthesis.LABEL_SHORTENING_STRATEGY);
  }
  
  @Inject
  @Extension
  private KPolylineExtensions _kPolylineExtensions;
  
  @Inject
  @Extension
  private KRenderingExtensions _kRenderingExtensions;
  
  @Inject
  @Extension
  private KColorExtensions _kColorExtensions;
  
  @Inject
  @Extension
  private KLibraryExtensions _kLibraryExtensions;
  
  /**
   * Rendering factory used to create KRendering model instances.
   */
  private static KRenderingFactory renderingFactory = KRenderingFactory.eINSTANCE;
  
  /**
   * The value for the defaults property as specified in the model.
   */
  private boolean defaults = false;
  
  /**
   * Default rendering for nodes.
   */
  private KRendering defaultNodeRendering;
  
  /**
   * Default rendering for polyline edges.
   */
  private KRendering defaultPolylineRendering;
  
  /**
   * Default rendering for spline edges.
   */
  private KRendering defaultSplineRendering;
  
  /**
   * Transforms the given graph into an equivalent graph that may be enriched with additional
   * rendering information.
   * 
   * @param graph the graph to transform.
   * @return the possibly enriched graph.
   */
  @Override
  public KNode transform(final KNode graph) {
    ViewContext _usedContext = this.getUsedContext();
    _usedContext.<Boolean>setProperty(KlighdSynthesisProperties.SUPPRESS_EDGE_ADJUSTMENT, Boolean.valueOf(true));
    final EcoreUtil.Copier copier = new EcoreUtil.Copier();
    EObject _copy = copier.copy(graph);
    final KNode result = ((KNode) _copy);
    copier.copyReferences();
    GraphDataUtil.loadDataElements(result, KGraphDiagramSynthesis.PREDICATE_IS_KGRAPHDATA, ((IProperty<?>[])Conversions.unwrapArray(KGraphDiagramSynthesis.KNOWN_PROPS, IProperty.class)));
    try {
      KLayoutData _data = graph.<KLayoutData>getData(KLayoutData.class);
      Boolean _property = _data.<Boolean>getProperty(KGraphDiagramSynthesis.DEFAULTS_PROPERTY);
      this.defaults = (_property).booleanValue();
    } catch (final Throwable _t) {
      if (_t instanceof ClassCastException) {
        final ClassCastException cce = (ClassCastException)_t;
      } else {
        throw Exceptions.sneakyThrow(_t);
      }
    }
    this.addLabelManager(result);
    KRenderingLibrary library = result.<KRenderingLibrary>getData(KRenderingLibrary.class);
    boolean _equals = Objects.equal(library, null);
    if (_equals) {
      KRenderingLibrary _createKRenderingLibrary = KGraphDiagramSynthesis.renderingFactory.createKRenderingLibrary();
      library = _createKRenderingLibrary;
      EList<KGraphData> _data_1 = result.getData();
      _data_1.add(library);
    }
    Object _objectValue = this.getObjectValue(KGraphDiagramSynthesis.STYLE);
    boolean _matched = false;
    if (!_matched) {
      if (Objects.equal(_objectValue, KGraphDiagramSynthesis.STYLE_STYLISH)) {
        _matched=true;
        this.initStylishFactory(library);
      }
    }
    if (!_matched) {
      if (Objects.equal(_objectValue, KGraphDiagramSynthesis.STYLE_HELLO_KITTY)) {
        _matched=true;
        this.initHelloKittyFactory(library);
      }
    }
    if (!_matched) {
      if (Objects.equal(_objectValue, KGraphDiagramSynthesis.STYLE_SPONGEBOB)) {
        _matched=true;
        this.initSpongebobFactory(library);
      }
    }
    if (!_matched) {
      this.initBoringFactory(library);
    }
    Set<Map.Entry<EObject, EObject>> _entrySet = copier.entrySet();
    for (final Map.Entry<EObject, EObject> entry : _entrySet) {
      EObject _value = entry.getValue();
      EObject _key = entry.getKey();
      this.<EObject>associateWith(_value, _key);
    }
    this.recursivelyEnrichRendering(result);
    return result;
  }
  
  /**
   * Configures the appropriate label manager for the given KGraph.
   * 
   * @param kgraph the graph to configure the label manager for.
   */
  public KNode addLabelManager(final KNode kgraph) {
    KNode _xblockexpression = null;
    {
      AbstractKlighdLabelManager labelManager = ((AbstractKlighdLabelManager) null);
      Object _objectValue = this.getObjectValue(KGraphDiagramSynthesis.LABEL_SHORTENING_STRATEGY);
      boolean _matched = false;
      if (!_matched) {
        if (Objects.equal(_objectValue, KGraphDiagramSynthesis.LABELS_NO_LABELS)) {
          _matched=true;
          java.util.function.Predicate<EObject> _matchNone = LabelPredicates.matchNone();
          ConditionLabelManager _conditionLabelManager = new ConditionLabelManager(
            null, _matchNone, 
            true);
          labelManager = _conditionLabelManager;
        }
      }
      if (!_matched) {
        if (Objects.equal(_objectValue, KGraphDiagramSynthesis.LABELS_TRUNCATE)) {
          _matched=true;
          TruncatingLabelManager _truncatingLabelManager = new TruncatingLabelManager();
          java.util.function.Predicate<EObject> _centerEdgeLabel = LabelPredicates.centerEdgeLabel();
          ConditionLabelManager _conditionLabelManager_1 = new ConditionLabelManager(_truncatingLabelManager, _centerEdgeLabel, 
            false);
          labelManager = _conditionLabelManager_1;
        }
      }
      if (!_matched) {
        if (Objects.equal(_objectValue, KGraphDiagramSynthesis.LABELS_SOFT_WORD_WRAP)) {
          _matched=true;
          SoftWrappingLabelManager _softWrappingLabelManager = new SoftWrappingLabelManager();
          java.util.function.Predicate<EObject> _centerEdgeLabel_1 = LabelPredicates.centerEdgeLabel();
          ConditionLabelManager _conditionLabelManager_2 = new ConditionLabelManager(_softWrappingLabelManager, _centerEdgeLabel_1, 
            false);
          labelManager = _conditionLabelManager_2;
        }
      }
      if (!_matched) {
        if (Objects.equal(_objectValue, KGraphDiagramSynthesis.LABELS_FULL)) {
          _matched=true;
          IdentLabelManager _identLabelManager = new IdentLabelManager();
          labelManager = _identLabelManager;
        }
      }
      KNode _xifexpression = null;
      boolean _notEquals = (!Objects.equal(labelManager, null));
      if (_notEquals) {
        _xifexpression = this.<KNode, ILabelManager>setLayoutOption(kgraph, LabelManagementOptions.LABEL_MANAGER, labelManager);
      }
      _xblockexpression = _xifexpression;
    }
    return _xblockexpression;
  }
  
  private boolean initEdgeRenderings(final KRenderingLibrary library) {
    boolean _xblockexpression = false;
    {
      KPolyline _createKPolyline = KGraphDiagramSynthesis.renderingFactory.createKPolyline();
      final Procedure1<KPolyline> _function = new Procedure1<KPolyline>() {
        @Override
        public void apply(final KPolyline it) {
          it.setId("DefaultEdgeRendering");
          KGraphDiagramSynthesis.this._kPolylineExtensions.addHeadArrowDecorator(it);
          KGraphDiagramSynthesis.this._kPolylineExtensions.addJunctionPointDecorator(it);
        }
      };
      KPolyline _doubleArrow = ObjectExtensions.<KPolyline>operator_doubleArrow(_createKPolyline, _function);
      this.defaultPolylineRendering = _doubleArrow;
      EList<KStyleHolder> _renderings = library.getRenderings();
      _renderings.add(this.defaultPolylineRendering);
      KSpline _createKSpline = KGraphDiagramSynthesis.renderingFactory.createKSpline();
      final Procedure1<KSpline> _function_1 = new Procedure1<KSpline>() {
        @Override
        public void apply(final KSpline it) {
          it.setId("SplineEdgeRendering");
          KGraphDiagramSynthesis.this._kPolylineExtensions.addHeadArrowDecorator(it);
        }
      };
      KSpline _doubleArrow_1 = ObjectExtensions.<KSpline>operator_doubleArrow(_createKSpline, _function_1);
      this.defaultSplineRendering = _doubleArrow_1;
      EList<KStyleHolder> _renderings_1 = library.getRenderings();
      _xblockexpression = _renderings_1.add(this.defaultSplineRendering);
    }
    return _xblockexpression;
  }
  
  private boolean initBoringFactory(final KRenderingLibrary library) {
    boolean _xblockexpression = false;
    {
      this.initEdgeRenderings(library);
      KRectangle _createKRectangle = KGraphDiagramSynthesis.renderingFactory.createKRectangle();
      final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
        @Override
        public void apply(final KRectangle it) {
          it.setId("DefaultNodeRendering");
        }
      };
      KRectangle _doubleArrow = ObjectExtensions.<KRectangle>operator_doubleArrow(_createKRectangle, _function);
      this.defaultNodeRendering = _doubleArrow;
      EList<KStyleHolder> _renderings = library.getRenderings();
      _xblockexpression = _renderings.add(this.defaultNodeRendering);
    }
    return _xblockexpression;
  }
  
  private boolean initStylishFactory(final KRenderingLibrary library) {
    boolean _xblockexpression = false;
    {
      this.initEdgeRenderings(library);
      KRoundedRectangle _createKRoundedRectangle = KGraphDiagramSynthesis.renderingFactory.createKRoundedRectangle();
      final Procedure1<KRoundedRectangle> _function = new Procedure1<KRoundedRectangle>() {
        @Override
        public void apply(final KRoundedRectangle it) {
          it.setId("DefaultNodeRendering");
          it.setCornerWidth(5);
          it.setCornerHeight(5);
          KColor _color = KGraphDiagramSynthesis.this._kColorExtensions.getColor("#EBF6FA");
          KColor _color_1 = KGraphDiagramSynthesis.this._kColorExtensions.getColor("#C4E3F3");
          KGraphDiagramSynthesis.this._kRenderingExtensions.<KRoundedRectangle>setBackgroundGradient(it, _color, _color_1, 90);
        }
      };
      KRoundedRectangle _doubleArrow = ObjectExtensions.<KRoundedRectangle>operator_doubleArrow(_createKRoundedRectangle, _function);
      this.defaultNodeRendering = _doubleArrow;
      EList<KStyleHolder> _renderings = library.getRenderings();
      _xblockexpression = _renderings.add(this.defaultNodeRendering);
    }
    return _xblockexpression;
  }
  
  private boolean initHelloKittyFactory(final KRenderingLibrary library) {
    boolean _xblockexpression = false;
    {
      this.initEdgeRenderings(library);
      KEllipse _createKEllipse = KGraphDiagramSynthesis.renderingFactory.createKEllipse();
      final Procedure1<KEllipse> _function = new Procedure1<KEllipse>() {
        @Override
        public void apply(final KEllipse it) {
          it.setId("DefaultNodeRendering");
          KColor _color = KGraphDiagramSynthesis.this._kColorExtensions.getColor("#FFEEEE");
          KColor _color_1 = KGraphDiagramSynthesis.this._kColorExtensions.getColor("#FFBBBB");
          KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setBackgroundGradient(it, _color, _color_1, 90);
        }
      };
      KEllipse _doubleArrow = ObjectExtensions.<KEllipse>operator_doubleArrow(_createKEllipse, _function);
      this.defaultNodeRendering = _doubleArrow;
      EList<KStyleHolder> _renderings = library.getRenderings();
      _xblockexpression = _renderings.add(this.defaultNodeRendering);
    }
    return _xblockexpression;
  }
  
  private boolean initSpongebobFactory(final KRenderingLibrary library) {
    boolean _xblockexpression = false;
    {
      this.initEdgeRenderings(library);
      KRoundedRectangle _createKRoundedRectangle = KGraphDiagramSynthesis.renderingFactory.createKRoundedRectangle();
      final Procedure1<KRoundedRectangle> _function = new Procedure1<KRoundedRectangle>() {
        @Override
        public void apply(final KRoundedRectangle it) {
          it.setId("DefaultNodeRendering");
          it.setCornerWidth(5);
          it.setCornerHeight(5);
          KGraphDiagramSynthesis.this._kRenderingExtensions.<KRoundedRectangle>setBackgroundColor(it, 237, 249, 107);
          EList<KRendering> _children = it.getChildren();
          KEllipse _createKEllipse = KGraphDiagramSynthesis.renderingFactory.createKEllipse();
          final Procedure1<KEllipse> _function = new Procedure1<KEllipse>() {
            @Override
            public void apply(final KEllipse it) {
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setPointPlacementData(it, PositionReferenceX.LEFT, 0, 0.1f, PositionReferenceY.TOP, 0, 0.1f, 
                HorizontalAlignment.LEFT, VerticalAlignment.TOP, 0, 0, 20, 20);
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setBackgroundColor(it, 255, 255, 255);
            }
          };
          KEllipse _doubleArrow = ObjectExtensions.<KEllipse>operator_doubleArrow(_createKEllipse, _function);
          _children.add(_doubleArrow);
          EList<KRendering> _children_1 = it.getChildren();
          KEllipse _createKEllipse_1 = KGraphDiagramSynthesis.renderingFactory.createKEllipse();
          final Procedure1<KEllipse> _function_1 = new Procedure1<KEllipse>() {
            @Override
            public void apply(final KEllipse it) {
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setPointPlacementData(it, PositionReferenceX.RIGHT, 0, 0.2f, PositionReferenceY.TOP, 0, 0.15f, 
                HorizontalAlignment.LEFT, VerticalAlignment.TOP, 0, 0, 10, 10);
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setBackgroundColor(it, 255, 255, 255);
            }
          };
          KEllipse _doubleArrow_1 = ObjectExtensions.<KEllipse>operator_doubleArrow(_createKEllipse_1, _function_1);
          _children_1.add(_doubleArrow_1);
          EList<KRendering> _children_2 = it.getChildren();
          KEllipse _createKEllipse_2 = KGraphDiagramSynthesis.renderingFactory.createKEllipse();
          final Procedure1<KEllipse> _function_2 = new Procedure1<KEllipse>() {
            @Override
            public void apply(final KEllipse it) {
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setPointPlacementData(it, PositionReferenceX.LEFT, 0, 0.05f, PositionReferenceY.TOP, 0, 0.5f, 
                HorizontalAlignment.LEFT, VerticalAlignment.TOP, 0, 0, 10, 10);
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setBackgroundColor(it, 255, 255, 255);
            }
          };
          KEllipse _doubleArrow_2 = ObjectExtensions.<KEllipse>operator_doubleArrow(_createKEllipse_2, _function_2);
          _children_2.add(_doubleArrow_2);
          EList<KRendering> _children_3 = it.getChildren();
          KEllipse _createKEllipse_3 = KGraphDiagramSynthesis.renderingFactory.createKEllipse();
          final Procedure1<KEllipse> _function_3 = new Procedure1<KEllipse>() {
            @Override
            public void apply(final KEllipse it) {
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setPointPlacementData(it, PositionReferenceX.RIGHT, 0, 0.4f, PositionReferenceY.TOP, 0, 0.5f, 
                HorizontalAlignment.LEFT, VerticalAlignment.TOP, 0, 0, 10, 10);
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setBackgroundColor(it, 255, 255, 255);
            }
          };
          KEllipse _doubleArrow_3 = ObjectExtensions.<KEllipse>operator_doubleArrow(_createKEllipse_3, _function_3);
          _children_3.add(_doubleArrow_3);
          EList<KRendering> _children_4 = it.getChildren();
          KEllipse _createKEllipse_4 = KGraphDiagramSynthesis.renderingFactory.createKEllipse();
          final Procedure1<KEllipse> _function_4 = new Procedure1<KEllipse>() {
            @Override
            public void apply(final KEllipse it) {
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setPointPlacementData(it, PositionReferenceX.LEFT, 0, 0.15f, PositionReferenceY.BOTTOM, 0, 0.3f, 
                HorizontalAlignment.LEFT, VerticalAlignment.TOP, 0, 0, 23, 23);
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setBackgroundColor(it, 255, 255, 255);
            }
          };
          KEllipse _doubleArrow_4 = ObjectExtensions.<KEllipse>operator_doubleArrow(_createKEllipse_4, _function_4);
          _children_4.add(_doubleArrow_4);
          EList<KRendering> _children_5 = it.getChildren();
          KEllipse _createKEllipse_5 = KGraphDiagramSynthesis.renderingFactory.createKEllipse();
          final Procedure1<KEllipse> _function_5 = new Procedure1<KEllipse>() {
            @Override
            public void apply(final KEllipse it) {
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setPointPlacementData(it, PositionReferenceX.RIGHT, 0, 0.3f, PositionReferenceY.BOTTOM, 0, 0.4f, 
                HorizontalAlignment.LEFT, VerticalAlignment.TOP, 0, 0, 20, 20);
              KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setBackgroundColor(it, 255, 255, 255);
            }
          };
          KEllipse _doubleArrow_5 = ObjectExtensions.<KEllipse>operator_doubleArrow(_createKEllipse_5, _function_5);
          _children_5.add(_doubleArrow_5);
        }
      };
      KRoundedRectangle _doubleArrow = ObjectExtensions.<KRoundedRectangle>operator_doubleArrow(_createKRoundedRectangle, _function);
      this.defaultNodeRendering = _doubleArrow;
      EList<KStyleHolder> _renderings = library.getRenderings();
      _xblockexpression = _renderings.add(this.defaultNodeRendering);
    }
    return _xblockexpression;
  }
  
  /**
   * Recursively calls {@link #enrichRendering(EObject)} on this object and all of its children.
   * 
   * @param o object to enrich.
   */
  private void recursivelyEnrichRendering(final EObject o) {
    this.enrichRendering(o);
    EList<EObject> _eContents = o.eContents();
    final Consumer<EObject> _function = new Consumer<EObject>() {
      @Override
      public void accept(final EObject child) {
        KGraphDiagramSynthesis.this.recursivelyEnrichRendering(child);
      }
    };
    _eContents.forEach(_function);
  }
  
  /**
   * Fallback metchod that does not add any rendering information to the given object.
   * 
   * @param o an EMF object.
   */
  private void _enrichRendering(final EObject o) {
  }
  
  /**
   * Make sure the node has a valid size and a label (if defaults are switched on).
   * 
   * @param node the node whose rendering to enrich.
   */
  private void _enrichRendering(final KNode node) {
    boolean _defaultsEnabled = this.defaultsEnabled();
    boolean _not = (!_defaultsEnabled);
    if (_not) {
      return;
    }
    ElkUtil.configureWithDefaultValues(node);
    final KRenderingRef rendering = this._kLibraryExtensions.addRenderingRef(node, this.defaultNodeRendering);
    this._kRenderingExtensions.<KRenderingRef>addSingleClickAction(rendering, FocusAndContextAction.ID);
  }
  
  /**
   * Make sure the port has a valid size and a label (if defaults are switched on).
   * 
   * @param port the port whose rendering to enrich.
   */
  private void _enrichRendering(final KPort port) {
    boolean _defaultsEnabled = this.defaultsEnabled();
    boolean _not = (!_defaultsEnabled);
    if (_not) {
      return;
    }
    ElkUtil.configureWithDefaultValues(port);
  }
  
  /**
   * Possibly adds a proper rendering to the given edge. If no rendering is associated with the edge
   * yet, a default rendering with an arrowhead is added to it.
   * 
   * @param edge the edge whose rendering to enrich.
   */
  private void _enrichRendering(final KEdge edge) {
    boolean _hasRendering = this.hasRendering(edge);
    boolean _not = (!_hasRendering);
    if (_not) {
      KNode _xifexpression = null;
      KNode _target = edge.getTarget();
      KNode _source = edge.getSource();
      boolean _isDescendant = ElkUtil.isDescendant(_target, _source);
      if (_isDescendant) {
        _xifexpression = edge.getSource();
      } else {
        KNode _source_1 = edge.getSource();
        KNode _parent = null;
        if (_source_1!=null) {
          _parent=_source_1.getParent();
        }
        _xifexpression = _parent;
      }
      final KNode parent = _xifexpression;
      KLayoutData _data = null;
      if (parent!=null) {
        _data=parent.<KLayoutData>getData(KLayoutData.class);
      }
      EdgeRouting _property = null;
      if (_data!=null) {
        _property=_data.<EdgeRouting>getProperty(CoreOptions.EDGE_ROUTING);
      }
      final EdgeRouting routing = _property;
      EList<KGraphData> _data_1 = edge.getData();
      KRenderingRef _createKRenderingRef = KGraphDiagramSynthesis.renderingFactory.createKRenderingRef();
      final Procedure1<KRenderingRef> _function = new Procedure1<KRenderingRef>() {
        @Override
        public void apply(final KRenderingRef it) {
          boolean _and = false;
          boolean _notEquals = (!Objects.equal(routing, null));
          if (!_notEquals) {
            _and = false;
          } else {
            boolean _equals = Objects.equal(routing, EdgeRouting.SPLINES);
            _and = _equals;
          }
          if (_and) {
            it.setRendering(KGraphDiagramSynthesis.this.defaultSplineRendering);
          } else {
            it.setRendering(KGraphDiagramSynthesis.this.defaultPolylineRendering);
          }
          KGraphDiagramSynthesis.this._kRenderingExtensions.<KRenderingRef>addSingleClickAction(it, FocusAndContextAction.ID);
        }
      };
      KRenderingRef _doubleArrow = ObjectExtensions.<KRenderingRef>operator_doubleArrow(_createKRenderingRef, _function);
      _data_1.add(_doubleArrow);
    }
    boolean _defaultsEnabled = this.defaultsEnabled();
    boolean _not_1 = (!_defaultsEnabled);
    if (_not_1) {
      return;
    }
    ElkUtil.configureWithDefaultValues(edge);
  }
  
  /**
   * Possibly adds a proper rendering to the given label. If no rendering is associated with the label
   * yet, a default text rendering with a predefined font size is added to it. Also, makes sure that
   * an edge label placement is set on the label if it belongs to an edge.
   * 
   * @param edge the edge whose rendering to enrich.
   */
  private void _enrichRendering(final KLabel label) {
    boolean _hasRendering = this.hasRendering(label);
    boolean _not = (!_hasRendering);
    if (_not) {
      KText _createKText = KGraphDiagramSynthesis.renderingFactory.createKText();
      final Procedure1<KText> _function = new Procedure1<KText>() {
        @Override
        public void apply(final KText text) {
          EList<KGraphData> _data = label.getData();
          _data.add(text);
          KGraphDiagramSynthesis.this._kRenderingExtensions.setFontSize(text, (KlighdConstants.DEFAULT_FONT_SIZE - 2));
          KGraphDiagramSynthesis.this._kRenderingExtensions.<KText>addSingleClickAction(text, FocusAndContextAction.ID);
          EObject _eContainer = label.eContainer();
          if ((_eContainer instanceof KPort)) {
            KGraphDiagramSynthesis.this._kRenderingExtensions.setFontSize(text, (KlighdConstants.DEFAULT_FONT_SIZE - 3));
          }
        }
      };
      ObjectExtensions.<KText>operator_doubleArrow(_createKText, _function);
    }
  }
  
  /**
   * Checks if the given graph element already has a rendering attached.
   * 
   * @param e the element to check.
   * @return {@code true} if the element already has a rendering attached, {@code false} otherwise.
   */
  private boolean hasRendering(final KGraphElement e) {
    EList<KGraphData> _data = e.getData();
    final Function1<KGraphData, Boolean> _function = new Function1<KGraphData, Boolean>() {
      @Override
      public Boolean apply(final KGraphData it) {
        return Boolean.valueOf((it instanceof KRendering));
      }
    };
    return IterableExtensions.<KGraphData>exists(_data, _function);
  }
  
  /**
   * Returns whether or not default stuff should be added or not.
   * 
   * @return {@code true} if default stuff should be added, {@code false} otherwise.
   */
  private boolean defaultsEnabled() {
    boolean _or = false;
    Object _objectValue = this.getObjectValue(KGraphDiagramSynthesis.DEFAULTS);
    boolean _equals = Objects.equal(_objectValue, KGraphDiagramSynthesis.DEFAULTS_ON);
    if (_equals) {
      _or = true;
    } else {
      boolean _and = false;
      Object _objectValue_1 = this.getObjectValue(KGraphDiagramSynthesis.DEFAULTS);
      boolean _equals_1 = Objects.equal(_objectValue_1, KGraphDiagramSynthesis.DEFAULTS_AS_IN_MODEL);
      if (!_equals_1) {
        _and = false;
      } else {
        _and = this.defaults;
      }
      _or = _and;
    }
    return _or;
  }
  
  private void enrichRendering(final EObject edge) {
    if (edge instanceof KEdge) {
      _enrichRendering((KEdge)edge);
      return;
    } else if (edge instanceof KNode) {
      _enrichRendering((KNode)edge);
      return;
    } else if (edge instanceof KPort) {
      _enrichRendering((KPort)edge);
      return;
    } else if (edge instanceof KLabel) {
      _enrichRendering((KLabel)edge);
      return;
    } else if (edge != null) {
      _enrichRendering(edge);
      return;
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(edge).toString());
    }
  }
}
