/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://rtsys.informatik.uni-kiel.de/kieler
 * 
 * Copyright 2013 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd;

import com.google.common.collect.ImmutableList;
import com.google.inject.Inject;
import de.cau.cs.kieler.klighd.SynthesisOption;
import de.cau.cs.kieler.klighd.labels.ConditionLabelManager;
import de.cau.cs.kieler.klighd.labels.LabelPredicates;
import de.cau.cs.kieler.klighd.syntheses.AbstractDiagramSynthesis;
import de.cau.cs.kieler.klighd.syntheses.DiagramSyntheses;
import de.cau.cs.kieler.ptolemy.klighd.PortLabelDisplayStyle;
import de.cau.cs.kieler.ptolemy.klighd.transformation.CommentsAttachor;
import de.cau.cs.kieler.ptolemy.klighd.transformation.CommentsExtractor;
import de.cau.cs.kieler.ptolemy.klighd.transformation.Ptolemy2KGraphOptimization;
import de.cau.cs.kieler.ptolemy.klighd.transformation.Ptolemy2KGraphTransformation;
import de.cau.cs.kieler.ptolemy.klighd.transformation.Ptolemy2KGraphVisualization;
import java.util.List;
import java.util.function.Predicate;
import org.eclipse.elk.alg.layered.p4nodes.NodePlacementStrategy;
import org.eclipse.elk.alg.layered.properties.LayeredOptions;
import org.eclipse.elk.core.labels.ILabelManager;
import org.eclipse.elk.core.labels.LabelManagementOptions;
import org.eclipse.elk.core.options.CoreOptions;
import org.eclipse.elk.core.util.Pair;
import org.eclipse.elk.graph.KNode;
import org.eclipse.elk.graph.properties.IProperty;
import org.eclipse.emf.ecore.EObject;
import org.ptolemy.moml.DocumentRoot;

/**
 * Synthesis for turning Ptolemy models into KGraphs.
 * 
 * @author cds
 */
@SuppressWarnings("all")
public class PtolemyDiagramSynthesis extends AbstractDiagramSynthesis<DocumentRoot> {
  /**
   * Container class for synthesis options.
   */
  public static final class Options {
    public boolean comments;
    
    public boolean relations;
    
    public PortLabelDisplayStyle portLabels;
    
    public boolean properties;
    
    public boolean directors;
    
    public boolean attachComments;
    
    public boolean flatten;
    
    public int compoundNodeAlpha;
    
    public Options(final PtolemyDiagramSynthesis s) {
      boolean _booleanValue = s.getBooleanValue(PtolemyDiagramSynthesis.SHOW_COMMENTS);
      this.comments = _booleanValue;
      boolean _booleanValue_1 = s.getBooleanValue(PtolemyDiagramSynthesis.SHOW_RELATIONS);
      this.relations = _booleanValue_1;
      Object _objectValue = s.getObjectValue(PtolemyDiagramSynthesis.SHOW_PORT_LABELS);
      String _string = _objectValue.toString();
      PortLabelDisplayStyle _fromDisplayString = PortLabelDisplayStyle.fromDisplayString(_string);
      this.portLabels = _fromDisplayString;
      boolean _booleanValue_2 = s.getBooleanValue(PtolemyDiagramSynthesis.SHOW_PROPERTIES);
      this.properties = _booleanValue_2;
      boolean _booleanValue_3 = s.getBooleanValue(PtolemyDiagramSynthesis.SHOW_DIRECTORS);
      this.directors = _booleanValue_3;
      boolean _booleanValue_4 = s.getBooleanValue(PtolemyDiagramSynthesis.COMMENT_ATTACHMENT_HEURISTIC);
      this.attachComments = _booleanValue_4;
      boolean _booleanValue_5 = s.getBooleanValue(PtolemyDiagramSynthesis.FLATTEN);
      this.flatten = _booleanValue_5;
      int _intValue = s.getIntValue(PtolemyDiagramSynthesis.COMPOUND_NODE_ALPHA);
      this.compoundNodeAlpha = _intValue;
    }
  }
  
  public final static String ID = "de.cau.cs.kieler.ptolemy.klighd.PtolemyDiagramSynthesis";
  
  public final static SynthesisOption SHOW_COMMENTS = SynthesisOption.createCheckOption(
    "Comments", Boolean.valueOf(true));
  
  public final static SynthesisOption SHOW_RELATIONS = SynthesisOption.createCheckOption(
    "Relations", Boolean.valueOf(false));
  
  public final static SynthesisOption SHOW_PORT_LABELS = SynthesisOption.createChoiceOption(
    "Show Port Labels", ImmutableList.<String>of(
    PortLabelDisplayStyle.ALL.toString(), 
    PortLabelDisplayStyle.SELECTED_NODE.toString(), 
    PortLabelDisplayStyle.NONE.toString()), 
    PortLabelDisplayStyle.NONE.toString());
  
  public final static SynthesisOption SHOW_PROPERTIES = SynthesisOption.createCheckOption(
    "Parameters", Boolean.valueOf(true));
  
  public final static SynthesisOption SHOW_DIRECTORS = SynthesisOption.createCheckOption(
    "Directors", Boolean.valueOf(true));
  
  public final static SynthesisOption COMMENT_ATTACHMENT_HEURISTIC = SynthesisOption.createCheckOption("Comment attachment heuristic", Boolean.valueOf(true));
  
  public final static SynthesisOption FLATTEN = SynthesisOption.createCheckOption(
    "Flatten Composite Actors", Boolean.valueOf(false));
  
  public final static SynthesisOption COMPOUND_NODE_ALPHA = SynthesisOption.<Float>createRangeOption(
    "Nested model darkness", Float.valueOf(0f), Float.valueOf(255f), Float.valueOf(30f));
  
  public final static SynthesisOption ATTACHMENT_HEURISTIC = SynthesisOption.createChoiceOption(
    "Attachment heuristic", 
    ImmutableList.<String>of("Smallest distance", "Comment alignment", "Find label name plain", 
      "Find label name", "Find label name w/o two attached"), "Smallest distance");
  
  @Inject
  private Ptolemy2KGraphTransformation transformation;
  
  @Inject
  private Ptolemy2KGraphOptimization optimization;
  
  @Inject
  private Ptolemy2KGraphVisualization visualization;
  
  @Inject
  private CommentsExtractor commentsExtractor;
  
  @Inject
  private CommentsAttachor commentsAttachor;
  
  @Override
  public KNode transform(final DocumentRoot model) {
    final PtolemyDiagramSynthesis.Options options = new PtolemyDiagramSynthesis.Options(this);
    final KNode kgraph = this.transformation.transform(model, this);
    CommentsExtractor _xifexpression = null;
    if (options.comments) {
      _xifexpression = this.commentsExtractor;
    } else {
      _xifexpression = null;
    }
    this.optimization.optimize(kgraph, options, _xifexpression, this);
    this.visualization.visualize(kgraph, options);
    if (options.attachComments) {
      this.commentsAttachor.attachComments(kgraph);
    }
    Object _objectValue = this.getObjectValue(PtolemyDiagramSynthesis.SHOW_PORT_LABELS);
    boolean _equals = _objectValue.equals("Selected Node");
    if (_equals) {
      Predicate<EObject> _portLabel = LabelPredicates.portLabel();
      Predicate<EObject> _negate = _portLabel.negate();
      final ConditionLabelManager labelManager = new ConditionLabelManager(
        null, _negate, true);
      this.<KNode, ILabelManager>setLayoutOption(kgraph, LabelManagementOptions.LABEL_MANAGER, labelManager);
    }
    return kgraph;
  }
  
  /**
   * Diagram options.
   */
  @Override
  public List<SynthesisOption> getDisplayedSynthesisOptions() {
    SynthesisOption _createSeparator = SynthesisOption.createSeparator("Comments");
    SynthesisOption _createSeparator_1 = SynthesisOption.createSeparator("Hierarchy");
    return ImmutableList.<SynthesisOption>of(
      PtolemyDiagramSynthesis.SHOW_RELATIONS, 
      PtolemyDiagramSynthesis.SHOW_PORT_LABELS, 
      PtolemyDiagramSynthesis.SHOW_DIRECTORS, 
      PtolemyDiagramSynthesis.SHOW_PROPERTIES, _createSeparator, 
      PtolemyDiagramSynthesis.SHOW_COMMENTS, 
      PtolemyDiagramSynthesis.COMMENT_ATTACHMENT_HEURISTIC, _createSeparator_1, 
      PtolemyDiagramSynthesis.FLATTEN, 
      PtolemyDiagramSynthesis.COMPOUND_NODE_ALPHA);
  }
  
  /**
   * Layout options.
   */
  @Override
  public List<Pair<IProperty<?>, List<?>>> getDisplayedLayoutOptions() {
    NodePlacementStrategy[] _values = NodePlacementStrategy.values();
    ImmutableList<NodePlacementStrategy> _copyOf = ImmutableList.<NodePlacementStrategy>copyOf(_values);
    Pair<IProperty<?>, List<?>> _specifyLayoutOption = DiagramSyntheses.specifyLayoutOption(LayeredOptions.NODE_PLACEMENT_STRATEGY, _copyOf);
    ImmutableList<Integer> _of = ImmutableList.<Integer>of(Integer.valueOf(0), Integer.valueOf(255));
    Pair<IProperty<?>, List<?>> _specifyLayoutOption_1 = DiagramSyntheses.specifyLayoutOption(CoreOptions.SPACING_NODE, _of);
    return ImmutableList.<Pair<IProperty<?>, List<?>>>of(_specifyLayoutOption, _specifyLayoutOption_1);
  }
}
