/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation;

import com.google.inject.Guice;
import com.google.inject.Inject;
import com.google.inject.Injector;
import de.cau.cs.kieler.klighd.krendering.KRendering;
import de.cau.cs.kieler.ptolemy.klighd.PtolemyProperties;
import de.cau.cs.kieler.ptolemy.klighd.transformation.KRenderingFigureProvider;
import de.cau.cs.kieler.ptolemy.klighd.transformation.comments.ExplicitPtolemyAttachmentProvider;
import de.cau.cs.kieler.ptolemy.klighd.transformation.comments.PtolemyBoundsProvider;
import de.cau.cs.kieler.ptolemy.klighd.transformation.comments.PtolemyTitleCommentFilter;
import de.cau.cs.kieler.ptolemy.klighd.transformation.comments.ReferencePreferringAttachmentDecider;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.AnnotationExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.LabelExtensions;
import java.util.Collection;
import java.util.function.Function;
import org.eclipse.elk.core.comments.CommentAttacher;
import org.eclipse.elk.core.comments.DistanceHeuristic;
import org.eclipse.elk.core.comments.IBoundsProvider;
import org.eclipse.elk.core.comments.NodeReferenceHeuristic;
import org.eclipse.elk.core.comments.SizeEligibilityFilter;
import org.eclipse.elk.core.comments.TextPrefixFilter;
import org.eclipse.elk.core.klayoutdata.KLayoutData;
import org.eclipse.elk.graph.KEdge;
import org.eclipse.elk.graph.KGraphData;
import org.eclipse.elk.graph.KNode;
import org.eclipse.emf.common.util.EList;
import org.eclipse.xtext.xbase.lib.Extension;

/**
 * Tries to infer the attachments between comment nodes and the elements they are supposed to describe.
 * There are two cases here:</p>
 * 
 * <ol>
 *   <li>The model contains explicit links between at least one comment and model element. In this case,
 *   we assume that the model developer added all the links he wants explicitly and only transform such
 *   links into the our way of representing them. We recognize explicit links through the fact that the
 *   {@link CommentsExtractor} has already turned them into edges.</li>
 * 
 *   <li>The model does not contain any such explicit link. In this case, a heuristic is applied to each
 *   comment, trying to find the entity it is probably referring to.
 * </ol>
 * 
 * <p>The heuristic can be applied in the first case as well. The explicit links will not be touched,
 * then; only unlinked comments will be considered for attachment.</p>
 * 
 * @see CommentsExtractor
 * @author cds
 */
@SuppressWarnings("all")
public class CommentsAttachor {
  @Inject
  @Extension
  private KRenderingFigureProvider _kRenderingFigureProvider;
  
  @Inject
  @Extension
  private AnnotationExtensions _annotationExtensions;
  
  @Inject
  @Extension
  private LabelExtensions _labelExtensions;
  
  /**
   * Maximum distance for two objects to be considered close enough for attachment.
   */
  private final double maxAttachmentDistance = 50;
  
  /**
   * Maximum attachment distance used by the node reference heuristic.
   */
  private final double maxAttachmentDistanceForReferenceHeuristic = 30;
  
  /**
   * The maximum area of a comment to still be considered attacheable.
   */
  private final double maxCommentArea = 62000;
  
  /**
   * Invokes KIML's comment attachment framework on the given graph.
   * 
   * @param graph the graph to perform comment attachment on.
   */
  public void attachComments(final KNode graph) {
    final Injector injector = Guice.createInjector();
    final PtolemyBoundsProvider basicBoundsProvider = injector.<PtolemyBoundsProvider>getInstance(PtolemyBoundsProvider.class);
    final IBoundsProvider boundsProvider = basicBoundsProvider.cached();
    final ExplicitPtolemyAttachmentProvider explicitAttachmentProvider = injector.<ExplicitPtolemyAttachmentProvider>getInstance(ExplicitPtolemyAttachmentProvider.class);
    final PtolemyTitleCommentFilter titleCommentFilter = injector.<PtolemyTitleCommentFilter>getInstance(PtolemyTitleCommentFilter.class);
    CommentAttacher _commentAttacher = new CommentAttacher();
    CommentAttacher _withBoundsProvider = _commentAttacher.withBoundsProvider(boundsProvider);
    CommentAttacher _withExplicitAttachmentProvider = _withBoundsProvider.withExplicitAttachmentProvider(explicitAttachmentProvider);
    ReferencePreferringAttachmentDecider _referencePreferringAttachmentDecider = new ReferencePreferringAttachmentDecider();
    CommentAttacher _withAttachmentDecider = _withExplicitAttachmentProvider.withAttachmentDecider(_referencePreferringAttachmentDecider);
    CommentAttacher _addEligibilityFilter = _withAttachmentDecider.addEligibilityFilter(titleCommentFilter);
    TextPrefixFilter _textPrefixFilter = new TextPrefixFilter();
    final Function<KNode, String> _function = new Function<KNode, String>() {
      @Override
      public String apply(final KNode c) {
        KLayoutData _layout = CommentsAttachor.this._annotationExtensions.getLayout(c);
        return _layout.<String>getProperty(PtolemyProperties.COMMENT_TEXT);
      }
    };
    TextPrefixFilter _withCommentTextProvider = _textPrefixFilter.withCommentTextProvider(_function);
    TextPrefixFilter _addPrefix = _withCommentTextProvider.addPrefix("Author");
    TextPrefixFilter _addPrefix_1 = _addPrefix.addPrefix("Demo created by");
    TextPrefixFilter _addPrefix_2 = _addPrefix_1.addPrefix("This model ");
    TextPrefixFilter _addPrefix_3 = _addPrefix_2.addPrefix("This submodel ");
    TextPrefixFilter _addPrefix_4 = _addPrefix_3.addPrefix("This example ");
    TextPrefixFilter _addPrefix_5 = _addPrefix_4.addPrefix("This demo ");
    TextPrefixFilter _addPrefix_6 = _addPrefix_5.addPrefix("Model of ");
    CommentAttacher _addEligibilityFilter_1 = _addEligibilityFilter.addEligibilityFilter(_addPrefix_6);
    SizeEligibilityFilter _sizeEligibilityFilter = new SizeEligibilityFilter();
    SizeEligibilityFilter _withBoundsProvider_1 = _sizeEligibilityFilter.withBoundsProvider(boundsProvider);
    SizeEligibilityFilter _withMaximumArea = _withBoundsProvider_1.withMaximumArea(this.maxCommentArea);
    CommentAttacher _addEligibilityFilter_2 = _addEligibilityFilter_1.addEligibilityFilter(_withMaximumArea);
    NodeReferenceHeuristic _nodeReferenceHeuristic = new NodeReferenceHeuristic();
    final Function<KNode, String> _function_1 = new Function<KNode, String>() {
      @Override
      public String apply(final KNode c) {
        KLayoutData _layout = CommentsAttachor.this._annotationExtensions.getLayout(c);
        return _layout.<String>getProperty(PtolemyProperties.COMMENT_TEXT);
      }
    };
    NodeReferenceHeuristic _withCommentTextProvider_1 = _nodeReferenceHeuristic.withCommentTextProvider(_function_1);
    final Function<KNode, String> _function_2 = new Function<KNode, String>() {
      @Override
      public String apply(final KNode n) {
        return CommentsAttachor.this._labelExtensions.getName(n);
      }
    };
    NodeReferenceHeuristic _withNodeNameProvider = _withCommentTextProvider_1.withNodeNameProvider(_function_2);
    NodeReferenceHeuristic _withBoundsProvider_2 = _withNodeNameProvider.withBoundsProvider(boundsProvider);
    NodeReferenceHeuristic _withMaximumAttachmentDistance = _withBoundsProvider_2.withMaximumAttachmentDistance(this.maxAttachmentDistanceForReferenceHeuristic);
    CommentAttacher _addHeuristic = _addEligibilityFilter_2.addHeuristic(_withMaximumAttachmentDistance);
    DistanceHeuristic _distanceHeuristic = new DistanceHeuristic();
    DistanceHeuristic _withBoundsProvider_3 = _distanceHeuristic.withBoundsProvider(boundsProvider);
    DistanceHeuristic _withMaximumAttachmentDistance_1 = _withBoundsProvider_3.withMaximumAttachmentDistance(this.maxAttachmentDistance);
    final CommentAttacher attacher = _addHeuristic.addHeuristic(_withMaximumAttachmentDistance_1);
    final Collection<KEdge> edges = attacher.attachComments(graph);
    for (final KEdge edge : edges) {
      {
        final KRendering edgeRendering = this._kRenderingFigureProvider.createCommentEdgeRendering(edge);
        EList<KGraphData> _data = edge.getData();
        _data.add(edgeRendering);
      }
    }
  }
}
