/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation;

import com.google.common.base.Objects;
import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.Iterables;
import com.google.common.collect.Multimap;
import com.google.inject.Inject;
import de.cau.cs.kieler.klighd.syntheses.AbstractDiagramSynthesis;
import de.cau.cs.kieler.ptolemy.klighd.PtolemyProperties;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.AnnotationExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MarkerExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.TransformationConstants;
import java.util.List;
import java.util.Map;
import org.eclipse.elk.core.klayoutdata.KLayoutData;
import org.eclipse.elk.core.util.ElkUtil;
import org.eclipse.elk.core.util.Pair;
import org.eclipse.elk.graph.KNode;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.FeatureMap;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.emf.ecore.xml.type.AnyType;
import org.eclipse.xtext.xbase.lib.Extension;
import org.ptolemy.moml.ConfigureType;
import org.ptolemy.moml.PropertyType;

/**
 * Extracts comments from the model and turns them into special comment nodes.
 * 
 * <p>In Ptolemy, certain annotations in a model are like comments in source code.
 * There are two ways how they can be represented in MOML:</p>
 * 
 * <ol>
 *   <li>Using a property of type "ptolemy.vergil.kernel.attributes.TextAttribute"
 *       with another property named "text", as follows:
 *     <pre>
 *        &lt;property
 *              name="Annotation2"
 *              class="ptolemy.vergil.kernel.attributes.TextAttribute"&gt;
 * 
 *              &lt;property
 *                      name="text"
 *                      class="ptolemy.kernel.util.StringAttribute"
 *                      value="This is my annotation's text."&gt;
 *              &lt;/property&gt;
 *              &lt;property
 *                      name="_location"
 *                      class="ptolemy.kernel.util.Location"
 *                      value="[140.0, 440.0]"&gt;
 *              &lt;/property&gt;
 *        &lt;/property&gt;
 *     </pre>
 *   </li>
 * 
 *   <li>Using a property of type "ptolemy.kernel.util.Attribute" with an SVG
 *       as its "_iconDescription" property, as follows:
 *     <pre>
 *        &lt;property
 *              name="annotation3"
 *              class="ptolemy.kernel.util.Attribute"&gt;
 * 
 *              &lt;property
 *                      name="_iconDescription"
 *                      class="ptolemy.kernel.util.SingletonConfigurableAttribute"&gt;
 * 
 *                      &lt;configure&gt;&lt;svg&gt;
 *                              &lt;text x="20" y="20" style="..."&gt;
 *                                      This is my annotation's text.
 *                              &lt;/text&gt;
 *                      &lt;/svg&gt;&lt;/configure&gt;
 *              &lt;/property&gt;
 *              &lt;property
 *                      name="_location"
 *                      class="ptolemy.kernel.util.Location"
 *                      value="[325.0, 10.0]"&gt;
 *              &lt;/property&gt;
 *        &lt;/property&gt;
 *     </pre>
 *   </li>
 * </ol>
 * 
 * <p>while the first version is straightforward, the latter version causes a whole
 * lot of problems. The {@code configure} element is a mixed
 * element, which means that it can contain anything, not just XML. However, it
 * does contain XML (usually an {@code svg} element and its children), which
 * disturbs the quiet peace of the parser. (which, in turn, disturbs my quiet peace.)
 * The {@code configure} element and its children are then dropped by the parser
 * during the transformation and are added to a list of unknown features. Recovering
 * such comments is one of the two tasks of this handler.</p>
 * 
 * <p>The second task is recognizing links between comments and model elements. Such
 * links indicate that a comment is providing additional information for a model
 * element, and should be preserved. In the MoML file, such a link is represented by
 * the following property:</p>
 * 
 * <pre>
 *   &lt;property
 *         name="_location"
 *         class="ptolemy.kernel.util.RelativeLocation"
 *         value="[-195.0, -80.0]"&gt;
 * 
 *         &lt;property
 *               name="relativeTo"
 *               class="ptolemy.kernel.util.StringAttribute"
 *               value="Const"/&gt;
 *         &lt;property
 *               name="relativeToElementName"
 *               class="ptolemy.kernel.util.StringAttribute"
 *               value="entity"/&gt;
 *   &lt;/property&gt;
 * </pre>
 * 
 * <p>That is, the annotation's {@code _location} property is extended by two additional
 * properties that define that the location is to be interpreted relative to a given
 * model element. So far, we only support linking comments to entities, but that could
 * well be changed.</p>
 * 
 * @see CommentsAttachor
 * @author cds
 */
@SuppressWarnings("all")
public class CommentsExtractor {
  /**
   * Marking nodes.
   */
  @Inject
  @Extension
  private AnnotationExtensions _annotationExtensions;
  
  /**
   * Marking nodes.
   */
  @Inject
  @Extension
  private MarkerExtensions _markerExtensions;
  
  /**
   * Extracts the comments attached as annotations to the given node and turns them into proper
   * nodes that are children of the given node.
   * 
   * @param root the root node.
   * @param diagramSynthesis the diagram synthesis that uses this class; used to map Ptolemy model
   *                         objects to the nodes created for them.
   * @return extracted comment nodes, to be passed to the attachment heuristic later.
   */
  public Multimap<KNode, KNode> extractComments(final KNode root, final AbstractDiagramSynthesis<?> diagramSynthesis) {
    final Multimap<KNode, KNode> createdCommentNodes = ArrayListMultimap.<KNode, KNode>create();
    List<PropertyType> _annotations = this._annotationExtensions.getAnnotations(root);
    for (final PropertyType annotation : _annotations) {
      String _elvis = null;
      String _class_ = annotation.getClass_();
      if (_class_ != null) {
        _elvis = _class_;
      } else {
        _elvis = "";
      }
      boolean _equals = _elvis.equals(TransformationConstants.ANNOTATION_TYPE_TEXT_ATTRIBUTE);
      if (_equals) {
        String _elvis_1 = null;
        String _annotationValue = this._annotationExtensions.getAnnotationValue(annotation, TransformationConstants.ANNOTATION_COMMENT_TEXT);
        if (_annotationValue != null) {
          _elvis_1 = _annotationValue;
        } else {
          _elvis_1 = "";
        }
        String _elvis_2 = null;
        String _annotationValue_1 = this._annotationExtensions.getAnnotationValue(annotation, TransformationConstants.ANNOTATION_FONT_SIZE);
        if (_annotationValue_1 != null) {
          _elvis_2 = _annotationValue_1;
        } else {
          _elvis_2 = "14";
        }
        final KNode commentNode = this.addCommentNode(root, _elvis_1, _elvis_2, createdCommentNodes);
        List<PropertyType> _annotations_1 = this._annotationExtensions.getAnnotations(commentNode);
        List<PropertyType> _annotations_2 = this._annotationExtensions.getAnnotations(annotation);
        Iterables.<PropertyType>addAll(_annotations_1, _annotations_2);
        diagramSynthesis.<KNode>associateWith(commentNode, annotation);
      } else {
        String _elvis_3 = null;
        String _class__1 = annotation.getClass_();
        if (_class__1 != null) {
          _elvis_3 = _class__1;
        } else {
          _elvis_3 = "";
        }
        boolean _equals_1 = _elvis_3.equals(TransformationConstants.ANNOTATION_TYPE_ATTRIBUTE);
        if (_equals_1) {
          final PropertyType iconDescription = this._annotationExtensions.getAnnotation(annotation, "_iconDescription");
          boolean _notEquals = (!Objects.equal(iconDescription, null));
          if (_notEquals) {
            final Pair<String, String> commentDetails = this.restoreCommentTextFromIconDescription(iconDescription);
            boolean _and = false;
            boolean _notEquals_1 = (!Objects.equal(commentDetails, null));
            if (!_notEquals_1) {
              _and = false;
            } else {
              String _first = commentDetails.getFirst();
              boolean _notEquals_2 = (!Objects.equal(_first, null));
              _and = _notEquals_2;
            }
            if (_and) {
              String _first_1 = commentDetails.getFirst();
              String _second = commentDetails.getSecond();
              String _extractFontSize = this.extractFontSize(_second);
              final KNode commentNode_1 = this.addCommentNode(root, _first_1, _extractFontSize, createdCommentNodes);
              List<PropertyType> _annotations_3 = this._annotationExtensions.getAnnotations(commentNode_1);
              List<PropertyType> _annotations_4 = this._annotationExtensions.getAnnotations(annotation);
              Iterables.<PropertyType>addAll(_annotations_3, _annotations_4);
              diagramSynthesis.<KNode>associateWith(commentNode_1, annotation);
            }
          }
        }
      }
    }
    EList<KNode> _children = root.getChildren();
    for (final KNode child : _children) {
      boolean _isMarkedAsTitleNode = this._markerExtensions.isMarkedAsTitleNode(child);
      if (_isMarkedAsTitleNode) {
        createdCommentNodes.put(root, child);
      }
    }
    EList<KNode> _children_1 = root.getChildren();
    for (final KNode child_1 : _children_1) {
      EList<KNode> _children_2 = child_1.getChildren();
      boolean _isEmpty = _children_2.isEmpty();
      boolean _not = (!_isEmpty);
      if (_not) {
        this.extractComments(child_1, diagramSynthesis);
      }
    }
    return createdCommentNodes;
  }
  
  /**
   * Extracts the font size of a comment that was saved as an SVG graphic.
   * 
   * @param style the value of the {@code style} attribute of an attribute.
   * @return the font size of the element.
   */
  private String extractFontSize(final String style) {
    final String pattern = "font-size(\\s*):(\\s*)(\\d+).*";
    return style.replaceAll(pattern, "$3");
  }
  
  /**
   * Tries to restore the text of a comment that was not saved as a regular comment attribute, but
   * as an SVG graphic.
   * 
   * @param iconDescription the {@code _iconDescription} attribute of the attribute that might be a
   *                        comment. This is where we will look for the SVG graphic in.
   * @return a pair consisting of the text and the style of the text, if present, or {@code null} if
   *         the text could not be restored.
   */
  private Pair<String, String> restoreCommentTextFromIconDescription(final PropertyType iconDescription) {
    Resource _eResource = iconDescription.eResource();
    boolean _not = (!(_eResource instanceof XMLResource));
    if (_not) {
      return null;
    }
    Resource _eResource_1 = iconDescription.eResource();
    final XMLResource xmlResource = ((XMLResource) _eResource_1);
    EList<ConfigureType> _configure = iconDescription.getConfigure();
    boolean _isEmpty = _configure.isEmpty();
    if (_isEmpty) {
      return null;
    }
    EList<ConfigureType> _configure_1 = iconDescription.getConfigure();
    final ConfigureType configureElement = _configure_1.get(0);
    Map<EObject, AnyType> _eObjectToExtensionMap = xmlResource.getEObjectToExtensionMap();
    final AnyType unknownFeature = _eObjectToExtensionMap.get(configureElement);
    boolean _equals = Objects.equal(unknownFeature, null);
    if (_equals) {
      return null;
    }
    FeatureMap _mixed = unknownFeature.getMixed();
    final AnyType svgElement = this.findUnknownFeature(_mixed, "svg");
    boolean _equals_1 = Objects.equal(svgElement, null);
    if (_equals_1) {
      return null;
    }
    FeatureMap _mixed_1 = svgElement.getMixed();
    final AnyType textElement = this.findUnknownFeature(_mixed_1, "text");
    boolean _equals_2 = Objects.equal(textElement, null);
    if (_equals_2) {
      return null;
    }
    FeatureMap _mixed_2 = textElement.getMixed();
    boolean _isEmpty_1 = _mixed_2.isEmpty();
    if (_isEmpty_1) {
      return null;
    }
    FeatureMap _mixed_3 = textElement.getMixed();
    FeatureMap.Entry _get = _mixed_3.get(0);
    final Object data = _get.getValue();
    if ((!(data instanceof String))) {
      return null;
    } else {
      final Pair<String, String> commentDetails = Pair.<String, String>create();
      commentDetails.setFirst(((String) data));
      for (int i = 0; (i < textElement.getAnyAttribute().size()); i++) {
        {
          FeatureMap _anyAttribute = textElement.getAnyAttribute();
          final FeatureMap.Entry attribute = _anyAttribute.get(i);
          EStructuralFeature _eStructuralFeature = attribute.getEStructuralFeature();
          String _name = _eStructuralFeature.getName();
          boolean _equals_3 = Objects.equal(_name, "style");
          if (_equals_3) {
            Object _value = attribute.getValue();
            commentDetails.setSecond(((String) _value));
          }
        }
      }
      return commentDetails;
    }
  }
  
  /**
   * Creates a comment node and adds it to the given parent node. Also adds it to the list of
   * created comment nodes.
   * 
   * @param parent parent of the new comment node.
   * @param text the comment text to be displayed by the comment node.
   * @param fontSize the font size of the comment's text.
   * @param createdCommentNodes map of created comment nodes this node will be placed in.
   * @return the created comment node.
   */
  private KNode addCommentNode(final KNode parent, final String text, final String fontSize, final Multimap<KNode, KNode> createdCommentNodes) {
    final KNode commentNode = ElkUtil.createInitializedNode();
    KLayoutData _layout = this._annotationExtensions.getLayout(commentNode);
    _layout.<String>setProperty(PtolemyProperties.COMMENT_TEXT, text);
    KLayoutData _layout_1 = this._annotationExtensions.getLayout(commentNode);
    int _parseInt = Integer.parseInt(fontSize);
    _layout_1.<Integer>setProperty(PtolemyProperties.COMMENT_FONT_SIZE, Integer.valueOf(_parseInt));
    this._markerExtensions.markAsComment(commentNode);
    EList<KNode> _children = parent.getChildren();
    _children.add(commentNode);
    createdCommentNodes.put(parent, commentNode);
    return commentNode;
  }
  
  /**
   * Checks if the feature map contains a feature of the given name and returns that.
   * 
   * @param features the feature map to search through.
   * @param name name of the feature to look for.
   * @return the feature or {@code null} if none could be found with that name.
   */
  private AnyType findUnknownFeature(final FeatureMap features, final String name) {
    for (final FeatureMap.Entry entry : features) {
      EStructuralFeature _eStructuralFeature = entry.getEStructuralFeature();
      String _name = _eStructuralFeature.getName();
      boolean _equals = _name.equals(name);
      if (_equals) {
        Object _value = entry.getValue();
        if ((_value instanceof AnyType)) {
          Object _value_1 = entry.getValue();
          return ((AnyType) _value_1);
        }
      }
    }
    return null;
  }
}
