/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2013 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation;

import com.google.common.base.Objects;
import com.google.inject.Inject;
import de.cau.cs.kieler.klighd.krendering.KImage;
import de.cau.cs.kieler.klighd.krendering.KPosition;
import de.cau.cs.kieler.klighd.krendering.KRendering;
import de.cau.cs.kieler.klighd.krendering.KRenderingFactory;
import de.cau.cs.kieler.klighd.krendering.extensions.KRenderingExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.GraphicsUtils;
import diva.canvas.CanvasUtilities;
import diva.canvas.Figure;
import diva.canvas.toolbox.ImageFigure;
import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.List;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.w3c.dom.Document;
import ptolemy.data.expr.XMLParser;
import ptolemy.kernel.Entity;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.ConfigurableAttribute;
import ptolemy.moml.test.TestIconLoader;
import ptolemy.vergil.icon.EditorIcon;

/**
 * Loads Ptolemy stuff. Set the Ptolemy entity to load icons for, run it in the AWT event queue thread,
 * and retrieve the result.
 */
@SuppressWarnings("all")
public final class ImageLoadWorker implements Runnable {
  /**
   * KRendering utility methods.
   */
  @Inject
  @Extension
  private KRenderingExtensions _kRenderingExtensions;
  
  /**
   * Factory used to instantiate KRendering classes.
   */
  private final KRenderingFactory renderingFactory = KRenderingFactory.eINSTANCE;
  
  /**
   * The entity whose icon to load.
   */
  private Entity entity = null;
  
  /**
   * EditorIcons we have loaded for the entity.
   */
  private List<EditorIcon> loadedIcons = null;
  
  /**
   * Whether we have already waited for the unscaled images to finish loading. If so, we only need
   * to wait for the scaled images to finish loading the next time around.
   */
  private boolean unscaledImagesLoaded = false;
  
  /**
   * The rendering resulting from the loading operations.
   */
  private KRendering result = null;
  
  /**
   * Sets the entity that this worker object is to load icons for.
   * 
   * @param newEntity the entity to load icons for.
   */
  public void reset(final Entity newEntity) {
    this.entity = newEntity;
    this.loadedIcons = null;
    this.result = null;
    this.unscaledImagesLoaded = false;
  }
  
  /**
   * Returns the KRendering representation of the loaded entity icon.
   * 
   * @return the resulting KRednering representation.
   */
  public KRendering getResult() {
    return this.result;
  }
  
  /**
   * Loads icons for the entity set previously. Must be executed in the AWT event queue thread.
   */
  @Override
  public void run() {
    boolean _equals = Objects.equal(this.loadedIcons, null);
    if (_equals) {
      List<EditorIcon> _loadIconsForEntity = this.loadIconsForEntity(this.entity);
      this.loadedIcons = _loadIconsForEntity;
      boolean _isEmpty = this.loadedIcons.isEmpty();
      if (_isEmpty) {
        final Document svgDocument = this.loadSvgForEntity(this.entity);
        final KRendering figure = GraphicsUtils.createFigureFromSvg(svgDocument);
        this.result = figure;
      } else {
      }
    } else {
      if (this.unscaledImagesLoaded) {
        EditorIcon _get = this.loadedIcons.get(0);
        GraphicsUtils.waitForImages(_get, false);
        EditorIcon _get_1 = this.loadedIcons.get(0);
        KRendering _createRenderingFromIcon = this.createRenderingFromIcon(_get_1);
        this.result = _createRenderingFromIcon;
      } else {
        EditorIcon _get_2 = this.loadedIcons.get(0);
        GraphicsUtils.waitForImages(_get_2, true);
        this.unscaledImagesLoaded = true;
      }
    }
  }
  
  /**
   * Turns an editor icon into a proper rendering.
   * 
   * @param icon the icon.
   * @return the KRendering representation of the icon.
   */
  private KRendering createRenderingFromIcon(final EditorIcon icon) {
    final Figure ptFigure = icon.createBackgroundFigure();
    final Image figureImage = this.toImage(ptFigure);
    final int width = figureImage.getWidth(null);
    final int height = figureImage.getHeight(null);
    final BufferedImage resizedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
    final Graphics2D graphics = resizedImage.createGraphics();
    graphics.setComposite(AlphaComposite.Src);
    graphics.setRenderingHint(RenderingHints.KEY_INTERPOLATION, 
      RenderingHints.VALUE_INTERPOLATION_BILINEAR);
    graphics.setRenderingHint(RenderingHints.KEY_RENDERING, 
      RenderingHints.VALUE_RENDER_QUALITY);
    graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
      RenderingHints.VALUE_ANTIALIAS_ON);
    graphics.drawImage(figureImage, 0, 0, null);
    graphics.dispose();
    final ImageData swtImage = GraphicsUtils.convertToSwt(resizedImage);
    final KImage kImage = this.renderingFactory.createKImage();
    kImage.setImageObject(swtImage);
    KPosition _createKPosition = this._kRenderingExtensions.createKPosition(this._kRenderingExtensions.LEFT, 0, 0, this._kRenderingExtensions.TOP, 0, 0);
    KPosition _createKPosition_1 = this._kRenderingExtensions.createKPosition(this._kRenderingExtensions.LEFT, swtImage.width, 0, this._kRenderingExtensions.TOP, swtImage.height, 0);
    this._kRenderingExtensions.<KImage>setAreaPlacementData(kImage, _createKPosition, _createKPosition_1);
    return kImage;
  }
  
  /**
   * Loads the EditorIcons for a given entity.
   * 
   * @param entity the entity.
   * @return list of editor icons or {@code null} if none could be loaded.
   */
  private List<EditorIcon> loadIconsForEntity(final Entity entity) {
    final TestIconLoader loader = new TestIconLoader();
    try {
      String _className = entity.getClassName();
      loader.loadIconForClass(_className, entity);
      return entity.<EditorIcon>attributeList(EditorIcon.class);
    } catch (final Throwable _t) {
      if (_t instanceof Exception) {
        final Exception e = (Exception)_t;
        return null;
      } else {
        throw Exceptions.sneakyThrow(_t);
      }
    }
  }
  
  /**
   * Loads an SVG description of the figure used to represent the given entity, if any.
   * 
   * @param entity the entity.
   * @return the SVG description, if any could be loaded without problems.
   */
  private Document loadSvgForEntity(final Entity entity) {
    Attribute _attribute = entity.getAttribute("_iconDescription");
    String svgString = ((ConfigurableAttribute) _attribute).getConfigureText();
    Document svgDocument = null;
    try {
      final XMLParser xmlParser = new XMLParser();
      String _repairString = GraphicsUtils.repairString(svgString);
      svgString = _repairString;
      Document _parser = xmlParser.parser(svgString);
      svgDocument = _parser;
      Document _repairSvg = GraphicsUtils.repairSvg(svgDocument);
      svgDocument = _repairSvg;
      return svgDocument;
    } catch (final Throwable _t) {
      if (_t instanceof Exception) {
        final Exception e = (Exception)_t;
        return null;
      } else {
        throw Exceptions.sneakyThrow(_t);
      }
    }
  }
  
  /**
   * Tries to turn the given Diva figure into an AWT image.
   * 
   * @param figure the Diva figure.
   * @return the AWT image or {@code null} if anything went wrong.
   */
  private Image toImage(final Figure figure) {
    if ((figure instanceof ImageFigure)) {
      final Image image = ((ImageFigure) figure).getImage();
      boolean _equals = Objects.equal(image, null);
      if (_equals) {
        return null;
      } else {
        int _width = image.getWidth(null);
        int _height = image.getHeight(null);
        return image.getScaledInstance(_width, _height, Image.SCALE_DEFAULT);
      }
    } else {
      final Rectangle2D bounds = figure.getBounds();
      double _width_1 = bounds.getWidth();
      double _height_1 = bounds.getHeight();
      final Rectangle2D.Double size = new Rectangle2D.Double(0, 0, _width_1, _height_1);
      final AffineTransform transform = CanvasUtilities.computeFitTransform(bounds, size);
      figure.transform(transform);
      double _width_2 = bounds.getWidth();
      double _height_2 = bounds.getHeight();
      final BufferedImage image_1 = new BufferedImage(
        ((int) _width_2), ((int) _height_2), BufferedImage.TYPE_4BYTE_ABGR);
      final Graphics2D graphics = image_1.createGraphics();
      graphics.setRenderingHint(
        RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
      Color _color = new Color(255, 255, 255, 255);
      graphics.setBackground(_color);
      double _width_3 = bounds.getWidth();
      double _height_3 = bounds.getHeight();
      graphics.clearRect(0, 0, ((int) _width_3), ((int) _height_3));
      figure.paint(graphics);
      return image_1;
    }
  }
}
