/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://rtsys.informatik.uni-kiel.de/kieler
 * 
 * Copyright 2015 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation.comments;

import com.google.common.base.Objects;
import com.google.inject.Inject;
import de.cau.cs.kieler.ptolemy.klighd.PtolemyProperties;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.AnnotationExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MarkerExtensions;
import org.eclipse.elk.core.comments.IEligibilityFilter;
import org.eclipse.elk.core.klayoutdata.KLayoutData;
import org.eclipse.elk.core.options.CoreOptions;
import org.eclipse.elk.graph.KNode;
import org.eclipse.emf.common.util.EList;
import org.eclipse.xtext.xbase.lib.Extension;

/**
 * Passes judgement on a comment's eligibility for attachment based on whether it is considered a title
 * comment or not. There are two ways to be considered a title comment. First, the comment is a type
 * of title comment in Ptolemy. And second, it has the largest font size of all comments. We only
 * expect the latter to appear on the graph's top level.
 * 
 * <p>
 * The largest font size thing probably deserves a bit more details. To be considered a title comment
 * based on the font size, a comment must match a few conditions. First, it must have the largest font
 * size. Second, it has to be the only comment with that font size. Third, it must not be the only
 * comment. And finally, its font size needs to be larger than the default font size.
 * </p>
 */
@SuppressWarnings("all")
public final class PtolemyTitleCommentFilter implements IEligibilityFilter {
  @Inject
  @Extension
  private AnnotationExtensions _annotationExtensions;
  
  @Inject
  @Extension
  private MarkerExtensions _markerExtensions;
  
  /**
   * The comment with the largest font size in the current graph.
   */
  private KNode largestFontSizeComment = null;
  
  /**
   * Whether to decide only based on the font size, even if a comment is marked as title.
   */
  private boolean decideBasedOnFontSizeOnly = false;
  
  /**
   * Once this method is called, the filter will disregard whether a comment is marked as title and
   * only regard comments as title based on their font size. This should really only be necessary
   * for evaluations.
   */
  public void decideBasedOnFontSizeOnly() {
    this.decideBasedOnFontSizeOnly = true;
  }
  
  /**
   * Returns the comment determined by the filter to be the title comment. Can only be non-null
   * between calls to {@code preprocess(...)} and {@code cleanup()}.
   */
  public KNode getChosenComment() {
    return this.largestFontSizeComment;
  }
  
  /**
   * {@inheritDoc}
   */
  @Override
  public void preprocess(final KNode graph, final boolean includeHierarchy) {
    int largestFontSize = (PtolemyProperties.COMMENT_FONT_SIZE.getDefault()).intValue();
    int numberOfComments = 0;
    EList<KNode> _children = graph.getChildren();
    for (final KNode node : _children) {
      KLayoutData _layout = this._annotationExtensions.getLayout(node);
      Boolean _property = _layout.<Boolean>getProperty(CoreOptions.COMMENT_BOX);
      if ((_property).booleanValue()) {
        numberOfComments++;
        KLayoutData _layout_1 = this._annotationExtensions.getLayout(node);
        final Integer fontSize = _layout_1.<Integer>getProperty(PtolemyProperties.COMMENT_FONT_SIZE);
        if (((fontSize).intValue() > largestFontSize)) {
          largestFontSize = (fontSize).intValue();
          this.largestFontSizeComment = node;
        } else {
          if (((fontSize).intValue() == largestFontSize)) {
            this.largestFontSizeComment = null;
          }
        }
      }
    }
    if ((numberOfComments == 1)) {
      this.largestFontSizeComment = null;
    }
  }
  
  /**
   * {@inheritDoc}
   */
  @Override
  public boolean eligibleForAttachment(final KNode comment) {
    boolean _and = false;
    boolean _notEquals = (!Objects.equal(comment, this.largestFontSizeComment));
    if (!_notEquals) {
      _and = false;
    } else {
      boolean _or = false;
      if (this.decideBasedOnFontSizeOnly) {
        _or = true;
      } else {
        boolean _isMarkedAsTitleNode = this._markerExtensions.isMarkedAsTitleNode(comment);
        boolean _not = (!_isMarkedAsTitleNode);
        _or = _not;
      }
      _and = _or;
    }
    return _and;
  }
  
  /**
   * {@inheritDoc}
   */
  @Override
  public void cleanup() {
    this.largestFontSizeComment = null;
  }
}
